import re
import os.path
import itertools
import operator
import sys
import time
import logging
import glob
import uuid
import datetime
import shutil
import json
import tzlocal

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    from backports.zoneinfo import ZoneInfo

import cv2
import yaml
import numpy as np
import pandas as pd
import dateutil.parser

try:
    import bloscpack
except ImportError:
    bloscpack = None

from .constants import DEVNULL, STORE_MD_FILENAME, STORE_LOCK_FILENAME, STORE_MD_KEY, \
    STORE_INDEX_FILENAME, EXTRA_DATA_FILE_EXTENSIONS, FRAME_MD as _FRAME_MD
from .util import ImageCodecProcessor, JsonCustomEncoder, FourCC, ensure_color, \
    ensure_grayscale, motif_extra_data_h5_to_df, motif_extra_data_json_to_df, motif_extra_data_h5_attrs, \
    get_local_timezone_zoneinfo
from .index import ImgStoreIndex


_VERBOSE_DEBUG_GETS = False
_VERBOSE_DEBUG_CHUNKS = False
_VERBOSE_VERY = False  # overrides the other and prints all logs to stdout


def _extract_store_metadata(full_path):
    with open(full_path, 'r') as f:
        allmd = yaml.load(f, Loader=yaml.SafeLoader)
    return allmd.pop(STORE_MD_KEY)


# note: frame_idx always refers to the frame_idx within the chunk
# whereas frame_index refers to the global frame_index from (0, frame_count]


class _ImgStore(object):
    _version = 2
    _supported_modes = ''

    FRAME_MD = _FRAME_MD

    # noinspection PyShadowingBuiltins
    def __init__(self, basedir, mode, imgshape=None, imgdtype=np.uint8, chunksize=None, metadata=None,
                 encoding=None, write_encode_encoding=None, format=None, index=None):
        if mode not in self._supported_modes:
            raise ValueError('mode not supported')

        if imgdtype is not None:
            # ensure this is a string
            imgdtype = np.dtype(imgdtype).name

        self._basedir = basedir
        self._mode = mode
        self._imgshape = ()
        self._imgdtype = ''
        self._chunksize = 0
        self._encoding = None
        self._format = None
        self._codec_proc = ImageCodecProcessor()  # used in read and write mode
        self._decode_image = None
        self._encode_image = None
        self._uuid = None

        self._metadata = {}
        self._user_metadata = {}

        self._tN = self._t0 = time.time()

        self._created_utc = self._timezone_local = None

        self.frame_min = np.nan
        self.frame_max = np.nan
        self.frame_number = np.nan
        self.frame_count = 0
        self.frame_time = np.nan

        # noinspection PyClassHasNoInit
        class _Log:

            @staticmethod
            def info(msg):
                print(msg)

            debug = info
            warn = info
            error = info

        self._log = logging.getLogger('imgstore')
        if _VERBOSE_VERY:
            _VERBOSE_DEBUG_GETS = _VERBOSE_DEBUG_CHUNKS = True
            self._log = _Log

        self._chunk_n = 0
        self._chunk_current_frame_idx = -1
        self._chunk_n_and_chunk_paths = None

        # file pointer and filename of a file which can be used to store additional data per frame
        # (this is only created if data is actually stored)
        self._extra_data_fp = self._extra_data_fn = None

        if mode == 'w':
            if None in (imgshape, imgdtype, chunksize, format):
                raise ValueError('imgshape, imgdtype, chunksize, format must not be None')
            self._frame_n = 0
            self._init_write(imgshape, imgdtype, chunksize, metadata, encoding, write_encode_encoding, format)
        elif mode == 'r':
            self._init_read()

            self._index = None
            if index is None:
                # as we need to read the chunks, populate the chunk cache now too (could be an expensive operation
                # if there are thousands of chunks)
                t0 = time.time()
                self._chunk_n_and_chunk_paths = self._find_chunks(chunk_numbers=None)
                self._log.debug('found %s chunks in in %fs' % (len(self._chunk_n_and_chunk_paths), time.time() - t0))

                self._index = ImgStoreIndex.new_from_chunks(self._chunk_n_and_chunk_paths)
            elif (index is not None) and isinstance(index, ImgStoreIndex):
                self._log.debug('using supplied index')
                self._index = index
            else:
                raise TypeError('index must be of type ImgStoreIndex')

            self._t0 = self._index.frame_time_min
            self._tN = self._index.frame_time_max
            self.frame_count = self._index.frame_count
            self.frame_min = self._index.frame_min
            self.frame_max = self._index.frame_max

            # reset to the start of the file and load the first chunk
            self._load_chunk(0)
            assert self._chunk_current_frame_idx == -1
            assert self._chunk_n == 0
            self.frame_number = np.nan  # we haven't read any frames yet

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _readability_check(self, smd_class, smd_version):
        class_name = getattr(self, 'class_name', self.__class__.__name__)
        if smd_class != class_name:
            raise ValueError('incompatible store, can_read:%r opened:%r' % (class_name, smd_class))
        if smd_version != self._version:
            raise ValueError('incompatible store version')
        return True

    def _init_read(self):
        fullpath = os.path.join(self._basedir, STORE_MD_FILENAME)
        with open(fullpath, 'r') as f:
            allmd = yaml.load(f, Loader=yaml.SafeLoader)
        smd = allmd.pop(STORE_MD_KEY)

        self._metadata = smd
        self._user_metadata.update(allmd)

        self._readability_check(smd_class=smd['class'],
                                smd_version=smd['version'])

        try:
            # noinspection PyShadowingNames
            uuid = smd['uuid']
        except KeyError:
            # noinspection PyShadowingNames
            uuid = None
        self._uuid = uuid

        self._imgshape = tuple(smd['imgshape'])
        self._imgdtype = smd['imgdtype']
        self._chunksize = int(smd['chunksize'])
        self._encoding = smd['encoding']
        self._format = smd['format']

        # synthesize a created_date from old format stores
        if 'created_utc' not in smd:
            self._log.info('old store detected. synthesizing created datetime / timezone')

            dt = tz = None
            # we don't know the local timezone, so assume it is local
            if 'timezone' in allmd:
                # noinspection PyBroadException
                try:
                    tz = ZoneInfo(allmd['timezone'])
                except Exception:
                    pass
            if tz is None:
                tz = get_local_timezone_zoneinfo()

            # first the filename
            m = re.match(r"""(.*)(20[\d]{6}_\d{6}).*""", os.path.basename(self._basedir))
            if m:
                name, datestr = m.groups()
                # ive always been careful to make the files named with the local time
                time_tuple = time.strptime(datestr, '%Y%m%d_%H%M%S')
                _dt = datetime.datetime(*(time_tuple[0:6])).replace(tzinfo=tz)
                dt = _dt.astimezone(datetime.timezone.utc)

            # then the modification time of the file
            if dt is None:
                # file modifications are local time
                ts = os.path.getmtime(fullpath)
                local_tz = datetime.datetime.now().astimezone().tzinfo
                dt = datetime.datetime.fromtimestamp(ts, tz=local_tz).astimezone(datetime.timezone.utc)

            self._created_utc = dt
            self._timezone_local = tz
        else:
            # ensure that created_utc always has the pytz.utc timezone object because fuck you python
            # and fuck you dateutil for having different UTC tz objects
            # https://github.com/dateutil/dateutil/issues/131
            _dt = dateutil.parser.parse(smd['created_utc'])
            self._log.debug('parsed created_utc: %s (from %r)' % (_dt.isoformat(), _dt))
            try:
                self._created_utc = _dt.astimezone(datetime.timezone.utc)  # aware object can be in any timezone
            except ValueError:  # naive
                self._created_utc = _dt.replace(tzinfo=datetime.timezone.utc)  # d must be in UTC
            self._timezone_local = ZoneInfo(smd['timezone_local'])

        # if encoding is unset, autoconvert is no-op
        self._codec_proc.set_default_code(self._encoding)
        self._decode_image = self._codec_proc.autoconvert

    def _init_write(self, imgshape, imgdtype, chunksize, metadata, encoding, write_encode_encoding, fmt):
        for e in (encoding, write_encode_encoding):
            if e:
                if not self._codec_proc.check_code(e):
                    raise ValueError('unsupported store image encoding: %s' % e)

        # if encoding is unset, autoconvert is no-op
        self._codec_proc.set_default_code(write_encode_encoding)
        self._encode_image = self._codec_proc.autoconvert

        write_imgshape = self._calculate_written_image_shape(imgshape, fmt)

        if write_encode_encoding:
            # as we always encode to color
            write_imgshape = [write_imgshape[0], write_imgshape[1], 3]

        if not os.path.exists(self._basedir):
            os.makedirs(self._basedir)

        self._imgshape = imgshape
        self._imgdtype = imgdtype
        self._chunksize = chunksize
        self._format = fmt

        self._uuid = uuid.uuid4().hex
        # because fuck you python that utcnow is naieve. kind of fixed in python >3.2
        self._created_utc = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)
        self._timezone_local = get_local_timezone_zoneinfo()

        store_md = {'imgshape': write_imgshape,
                    'imgdtype': self._imgdtype,
                    'chunksize': chunksize,
                    'format': fmt,
                    'class': self.__class__.__name__,
                    'version': self._version,
                    'encoding': encoding,
                    # actually write the string as naieve because we have guarenteed it is UTC
                    'created_utc': self._created_utc.replace(tzinfo=None).isoformat(),
                    'timezone_local': str(self._timezone_local),
                    'uuid': self._uuid}

        if metadata is None:
            metadata = {STORE_MD_KEY: store_md}
        elif isinstance(metadata, dict):
            try:
                metadata[STORE_MD_KEY].update(store_md)
            except KeyError:
                metadata[STORE_MD_KEY] = store_md
        else:
            raise ValueError('metadata must be a dictionary')

        with open(os.path.join(self._basedir, STORE_MD_FILENAME), 'wt') as f:
            yaml.safe_dump(metadata, f)

        with open(os.path.join(self._basedir, STORE_LOCK_FILENAME), 'a') as _:
            pass

        smd = metadata.pop(STORE_MD_KEY)
        self._metadata = smd
        self._user_metadata.update(metadata)

        self._save_chunk(None, self._chunk_n)

    def _save_chunk_metadata(self, path_without_extension, extension='.npz'):
        path = path_without_extension + extension
        self._save_index(path, self._chunk_md)

        # also calculate the filename of the extra file to hold any data
        if self._extra_data_fp is not None:
            self._extra_data_fp.write(']')
            self._extra_data_fp.close()
            self._extra_data_fp = None

    def _new_chunk_metadata(self, chunk_path):
        self._extra_data_fn = chunk_path + '.extra.json'
        self._chunk_md = {k: [] for k in _ImgStore.FRAME_MD}
        self._chunk_md.update(self._metadata)

    def _save_image_metadata(self, frame_number, frame_time):
        self._chunk_md['frame_number'].append(frame_number)
        self._chunk_md['frame_time'].append(frame_time)

    # noinspection PyShadowingBuiltins,PyMethodMayBeStatic
    def _calculate_written_image_shape(self, imgshape, fmt):
        return imgshape

    def _iter_chunk_n_and_chunk_paths(self):
        if self._chunk_n_and_chunk_paths is None:
            t0 = time.time()
            self._chunk_n_and_chunk_paths = self._find_chunks(chunk_numbers=None)
            self._log.debug('found %s chunks in in %fs' % (len(self._chunk_n_and_chunk_paths), time.time() - t0))

        for cn, cp in sorted(self._chunk_n_and_chunk_paths, key=operator.itemgetter(0)):
            yield cn, cp

    @classmethod
    def supported_formats(cls):
        raise NotImplementedError

    @classmethod
    def supports_format(cls, fmt):
        return fmt in cls.supported_formats()

    @staticmethod
    def _save_index(path_with_extension, data_dict):
        _, extension = os.path.splitext(path_with_extension)
        if extension == '.yaml':
            with open(path_with_extension, 'wt') as f:
                yaml.safe_dump(data_dict, f)
                return path_with_extension
        elif extension == '.npz':
            with open(path_with_extension, 'wb') as f:
                # noinspection PyTypeChecker
                np.savez(f, **data_dict)
                return path_with_extension
        else:
            raise ValueError('unknown index format: %s' % extension)

    @staticmethod
    def _remove_index(path_without_extension):
        for extension in ('.npz', '.yaml'):
            path = path_without_extension + extension
            if os.path.exists(path):
                os.unlink(path)
                return path

    def _get_chunk_metadata(self, chunk_n):
        return self._index.get_chunk_metadata(chunk_n)

    def get_frame_metadata(self):
        return self._index.get_all_metadata()

    def _iter_extra_data_files(self, extensions=EXTRA_DATA_FILE_EXTENSIONS):
        for chunk_n, chunk_path in self._iter_chunk_n_and_chunk_paths():
            out = None

            # motif out of process IO puts extra data in the root directory
            for ext in ('.extra_data.h5', '.extra_data.json'):

                if ext not in extensions:
                    continue

                path = os.path.join(self._basedir, '%06d%s' % (int(chunk_n), ext))
                if os.path.exists(path):
                    yield True, path
                    break

            else:
                # for .. else
                #
                # if we _didn't_ break above then we enter here

                # imgstore API puts them beside the index,
                # which is a subdir for dirimgstores
                for ext in ('.extra.json', '.extra_data.json'):

                    if ext not in extensions:
                        continue

                    path = chunk_path + ext
                    if os.path.exists(path):
                        yield False, path

    @property
    def has_extra_data(self):
        for _ in self._iter_extra_data_files():
            return True
        return False

    def find_extra_data_files(self, extensions=EXTRA_DATA_FILE_EXTENSIONS):
        return [path for _, path in self._iter_extra_data_files(extensions=extensions)]

    def get_extra_data_samplerate(self):
        for c in self.find_extra_data_files(extensions=('.extra_data.h5',)):
            attrs = motif_extra_data_h5_attrs(c)
            ds = attrs['_datasets']
            if len(ds) > 1:
                raise ValueError('Multiple sample rates detected: please use motif_extra_data_h5_attrs')
            elif len(ds) == 1:
                try:
                    return attrs[ds[0]]['samplerate']
                except KeyError:
                    pass

        raise ValueError("No sample rate or dataset found in file")

    def get_extra_data(self, ignore_corrupt_chunks=False):
        dfs = []

        for is_motif, path in self._iter_extra_data_files():

            self._log.debug('found extra data chunk: %s (motif: %s)' % (path, is_motif))

            # noinspection PyBroadException
            try:
                if is_motif:
                    if path.endswith('.h5'):
                        df = motif_extra_data_h5_to_df(self, path)
                    elif path.endswith('.json'):
                        df = motif_extra_data_json_to_df(self, path)
                    else:
                        df = None

                    if (df is not None) and (not df.empty):
                        dfs.append(df)
                    else:
                        self._log.warn('chunk extra data %s is empty' % path)

                else:
                    if path.endswith('.json'):
                        with open(path, 'rt') as f:
                            records = json.load(f)
                        dfs.append(pd.DataFrame(records))
                    else:
                        raise ValueError('unknown extra data file encountered')

            except Exception:
                if ignore_corrupt_chunks:
                    self._log.warn('chunk extra data %s is corrupt' % path, exc_info=True)
                    continue
                else:
                    raise
        if dfs:
            return pd.concat(dfs, axis=0, ignore_index=True)

    def add_extra_data(self, **data):
        if not data:
            return

        data['frame_time'] = self.frame_time
        data['frame_number'] = self.frame_number
        data['frame_index'] = self._frame_n - 1  # we post-increment in add_frame

        # noinspection PyBroadException
        try:
            txt = json.dumps(data, cls=JsonCustomEncoder)
        except Exception:
            self._log.warn('error writing extra data', exc_info=True)
            return

        if self._extra_data_fp is None:
            self._extra_data_fp = open(self._extra_data_fn, 'w')
            self._extra_data_fp.write('[')
        else:
            self._extra_data_fp.write(', ')
        self._extra_data_fp.write(txt)

    def _save_image(self, img, frame_number, frame_time):  # pragma: no cover
        raise NotImplementedError

    def _save_chunk(self, old, new):  # pragma: no cover
        raise NotImplementedError

    def _load_image(self, idx):  # pragma: no cover
        raise NotImplementedError

    def _load_chunk(self, n):  # pragma: no cover
        raise NotImplementedError

    def _find_chunks(self, chunk_numbers):  # pragma: no cover
        raise NotImplementedError

    def __len__(self):
        return self.frame_count

    @property
    def created(self):
        return self._created_utc, self._timezone_local

    @property
    def uuid(self):
        return self._uuid

    @property
    def chunks(self):
        """ the number of non-empty chunks that contain images """
        if self._mode == 'r':
            return list(self._index.chunks)
        else:
            return list(range(0, self._chunk_n))

    @property
    def user_metadata(self):
        return self._user_metadata

    @property
    def filename(self):
        return self._basedir

    @property
    def full_path(self):
        return os.path.join(self._basedir, STORE_MD_FILENAME)

    @property
    def image_shape(self):
        # if encoding is specified, we always decode to bgr (color)
        if self._encoding:
            return self._imgshape[0], self._imgshape[1], 3
        else:
            return self._imgshape

    @property
    def duration(self):
        if np.isreal(self._tN) and np.isreal(self._t0):
            return self._tN - self._t0
        return 0.

    @property
    def frame_time_min(self):
        if np.isreal(self._t0):
            return self._t0

    @property
    def frame_time_max(self):
        if np.isreal(self._tN):
            return self._tN

    @property
    def mode(self):
        return self._mode

    @staticmethod
    def _extract_only_frame(basedir, chunk_n, frame_n, smd):
        raise NotImplementedError

    @classmethod
    def extract_only_frame(cls, full_path, frame_index, _smd=None):
        if _smd is None:
            smd = _extract_store_metadata(full_path)
        else:
            smd = _smd

        chunksize = int(smd['chunksize'])

        # go directly to the chunk
        chunk_n = frame_index // chunksize
        frame_n = frame_index % chunksize

        return cls._extract_only_frame(basedir=os.path.dirname(full_path),
                                       chunk_n=chunk_n,
                                       frame_n=frame_n,
                                       smd=smd)

    def convert_frame_time_to_datetime(self, frame_time, tz=None):
        """ converts a raw frame_time (usually seconds since the epoch), to a
        datetime object in the timezone of the imgstore, or that supplied (tz) """
        if tz is None:
            _, tz = self.created
        return datetime.datetime.fromtimestamp(frame_time, tz=tz)

    def disable_decoding(self):
        self._decode_image = lambda x: x

    def add_image(self, img, frame_number, frame_time):
        self._save_image(self._encode_image(img), frame_number, frame_time)

        self.frame_max = np.nanmax((frame_number, self.frame_max))
        self.frame_min = np.nanmin((frame_number, self.frame_min))
        self.frame_number = frame_number
        self.frame_time = frame_time

        if self._frame_n == 0:
            self._t0 = frame_time
        self._tN = frame_time

        self._frame_n += 1
        if (self._frame_n % self._chunksize) == 0:
            old = self._chunk_n
            new = self._chunk_n + 1
            self._save_chunk(old, new)
            self._chunk_n = new

        self.frame_count = self._frame_n

    def _get_next_framenumber_and_chunk_frame_idx(self):
        idx = self._chunk_current_frame_idx + 1

        try:
            # try fast to see if next frame is in currently loaded chunk
            frame_number = self._chunk_md['frame_number'][idx]
        except IndexError:
            # open the next chunk
            next_chunk = self._chunk_n + 1
            if next_chunk not in self._index.chunks:
                raise EOFError

            self._load_chunk(next_chunk)

            # first frame is start of chunk
            idx = 0
            frame_number = self._chunk_md['frame_number'][idx]

        return frame_number, idx

    def get_next_framenumber(self):
        return self._get_next_framenumber_and_chunk_frame_idx()[0]

    def _get_image(self, chunk_n, frame_idx):
        if chunk_n is not None:
            self._load_chunk(chunk_n)

        # ensure the read works before setting frame_number
        _img, (_frame_number, _frame_timestamp) = self._load_image(frame_idx)
        img = self._decode_image(_img)
        self._chunk_current_frame_idx = frame_idx
        self.frame_number = _frame_number

        return img, (_frame_number, _frame_timestamp)

    def get_next_image(self):
        frame_number, idx = self._get_next_framenumber_and_chunk_frame_idx()
        assert idx is not None

        if _VERBOSE_DEBUG_GETS:
            self._log.debug('get_next_image frame_number: %s idx %s' % (frame_number, idx))

        return self._get_image(chunk_n=None,  # chunk was already loaded in _get_next_framenumber_and_chunk_frame_idx
                               frame_idx=idx)

    def _get_image_by_frame_index(self, frame_index):
        """
        return the frame at the following index in the store
        """
        if frame_index < 0:
            raise ValueError('seeking to negative index not supported')

        if _VERBOSE_DEBUG_CHUNKS:
            self._log.debug('seek by frame_index %s' % frame_index)

        chunk_n, frame_idx = self._index.find_chunk('index', frame_index)

        if chunk_n == -1:
            raise ValueError('frame_index %s not found in index' % frame_index)

        self._log.debug('seek found in chunk %d attempt read chunk index %d' % (self._chunk_n, frame_idx))

        return self._get_image(chunk_n, frame_idx)

    def _get_image_by_frame_number(self, frame_number, exact_only):
        if _VERBOSE_DEBUG_CHUNKS:
            self._log.debug('seek by frame_number %s (exact: %s)' % (frame_number, exact_only))

        if exact_only:
            chunk_n, frame_idx = self._index.find_chunk('frame_number', frame_number)
        else:
            chunk_n, frame_idx = self._index.find_chunk_nearest('frame_number', frame_number)

        if chunk_n == -1:
            raise ValueError('frame #%s not found in any chunk' % frame_number)

        return self._get_image(chunk_n, frame_idx)

    def get_nearest_image(self, frame_time):
        chunk_n, frame_idx = self._index.find_chunk_nearest('frame_time', frame_time)
        return self._get_image(chunk_n, frame_idx)

    def get_image(self, frame_number, exact_only=True, frame_index=None):
        """
        seek to the supplied frame_number or frame_index. If frame_index is supplied get that image,
        otherwise get the image corresponding to frame_number

        :param frame_number:  (frame_min, frame_max)
        :param exact_only: If False return the nearest frame
        :param frame_index: frame_index (0, frame_count]
        """
        if _VERBOSE_DEBUG_GETS:
            self._log.debug('get_image %s (exact: %s) frame_idx %s' % (frame_number, exact_only, frame_index))
        if frame_index is not None:
            return self._get_image_by_frame_index(frame_index)
        else:
            return self._get_image_by_frame_number(frame_number, exact_only=exact_only)

    def close(self, save_index=False):
        if self._mode in 'wa':
            self._save_chunk(self._chunk_n, None)
            # noinspection PyBroadException
            try:
                if os.path.isfile(os.path.join(self._basedir, STORE_LOCK_FILENAME)):
                    os.remove(os.path.join(self._basedir, STORE_LOCK_FILENAME))
            except OSError:
                # noinspection PyArgumentList
                self._log.warn('could not remove lock file', exc_info=True)
            except Exception:
                # noinspection PyArgumentList
                self._log.warn('could not remove lock file (unknown error)', exc_info=True)

            if save_index:
                index = ImgStoreIndex.new_from_chunks(self._find_chunks(chunk_numbers=None))
                index.to_file(os.path.join(self._basedir, STORE_INDEX_FILENAME))

    def save_index(self, path=None):
        if self._mode == 'r':
            if path is None:
                path = os.path.join(self._basedir, STORE_INDEX_FILENAME)
            self._index.to_file(path)

    def empty(self):
        if self._mode != 'w':
            raise ValueError('can only empty stores for writing')

        self.close()

        self._tN = self._t0 = time.time()

        self.frame_min = np.nan
        self.frame_max = np.nan
        self.frame_number = np.nan
        self.frame_count = 0
        self.frame_time = np.nan

        self._chunk_n = 0
        self._chunk_n_and_chunk_paths = None

        if self._extra_data_fp is not None:
            self._extra_data_fp.close()
            os.unlink(self._extra_data_fn)
            self._extra_data_fp = self._extra_data_fn = None

        self._frame_n = 0

    def reindex(self):
        """ modifies the current imgstore so that all framenumbers before frame_number=0 are negative

        if there are multiple frame_numbers equal to zero then this operation aborts. this functions also
        updates the frame_number of any stored extra data. the original frame_number prior to calling
        reindex() is stored in '_frame_number_before_reindex' """
        md = self.get_frame_metadata()
        fn = md['frame_number']

        nzeros = fn.count(0)
        if nzeros != 1:
            raise ValueError("%d frame_number=0 found (should be 1)" % nzeros)

        # get index and time of sync frame
        zero_idx = fn.index(0)
        self._log.info('reindexing about frame_number=0 at index=%d' % zero_idx)

        fn_new = fn[:]
        fn_new[:zero_idx] = range(-zero_idx, 0)

        for chunk_n, chunk_path in self._iter_chunk_n_and_chunk_paths():
            ind = self._index.get_chunk_metadata(chunk_n)

            oft = list(ind['frame_time'])

            ofn = list(ind['frame_number'])
            nfn = fn_new[chunk_n*self._chunksize:(chunk_n*self._chunksize) + self._chunksize]
            assert len(ofn) == len(nfn)

            new_ind = {'frame_time': oft,
                       'frame_number': nfn,
                       '_frame_number_before_reindex': ofn}

            _path = self._remove_index(chunk_path)
            self._log.debug('reindex chunk %s removed: %s' % (chunk_n, _path))
            _path = self._save_index(chunk_path + '.npz', new_ind)
            self._log.debug('reindex chunk %s wrote: %s' % (chunk_n, _path))

            for ext in ('.extra.json', '.extra_data.json'):
                ed_path = chunk_path + ext
                if os.path.exists(ed_path):
                    with open(ed_path, 'r') as f:
                        ed = json.load(f)

                    # noinspection PyBroadException
                    try:
                        df = pd.DataFrame(ed)
                        if 'frame_index' not in df.columns:
                            raise ValueError('can not reindex extra-data on old format stores')
                        df['_frame_number_before_reindex'] = df['frame_number']
                        df['frame_number'] = df.apply(lambda r: fn_new[int(r.frame_index)], axis=1)
                        with open(ed_path, 'w') as f:
                            df.to_json(f, orient='records')
                        self._log.debug('reindexed chunk %d metadata (%s)' % (chunk_n, ed_path))
                    except Exception:
                        self._log.error('could not update chunk extra data to new framenumber', exc_info=True)

    @staticmethod
    def _extract_only_frame(basedir, chunk_n, frame_n, smd):
        return None


def _extract_ranges(data):
    # convert a list of integers into a list of contiguous ranges
    # [2, 3, 4, 5, 12, 13, 14, 15, 16, 17, 20] -> [(2,5), (12,17), (20, 20)]
    # http://stackoverflow.com/a/2154437
    ranges = []
    for key, group in itertools.groupby(enumerate(sorted(data)), lambda x: x[0] - x[1]):
        group = list(map(operator.itemgetter(1), group))
        if len(group) > 1:
            ranges.append((group[0], group[-1]))
        else:
            ranges.append((group[0], group[0]))

    return ranges


class DirectoryImgStore(_ImgStore):
    _supported_modes = 'wr'

    _cv2_fmts = {'tif', 'png', 'jpg', 'ppm', 'pgm', 'bmp', 'tif+color', 'png+color', 'jpg+color', 'bmp+color'}
    _raw_fmts = {'npy', 'bpk'}

    _DEFAULT_CHUNKSIZE = 200

    def __init__(self, **kwargs):

        self._chunk_cdir = ''
        self._chunk_md = {}

        # keep compat with VideoImgStore
        kwargs.pop('videosize', None)
        # keep compat with VideoImgStoreFFMPEG
        kwargs.pop('seek', None)
        kwargs.pop('gpu_id', None)

        if kwargs['mode'] == 'w':
            if 'chunksize' not in kwargs:
                kwargs['chunksize'] = self._DEFAULT_CHUNKSIZE
            kwargs['encoding'] = kwargs.pop('encoding', None)

        kwargs['write_encode_encoding'] = None

        if kwargs.get('format', '').endswith('+color'):
            kwargs['format'] = kwargs['format'].replace('+color', '')

            if kwargs['encoding']:
                kwargs['write_encode_encoding'] = kwargs['encoding']
                kwargs['encoding'] = None  # and so in the store it is no longer encoded

        _ImgStore.__init__(self, **kwargs)

        self._color = (self._imgshape[-1] == 3) & (len(self._imgshape) == 3)

        if (self._mode == 'w') and (self._format == 'pgm') and self._color:
            self._log.warn("store created with color image shape but using grayscale 'pgm' format")

        if self._format not in itertools.chain(self._cv2_fmts, ('npy', 'bpk')):
            raise ValueError('unknown format %s' % self._format)

        if (self._format == 'bpk') and (bloscpack is None):
            raise ValueError('bloscpack not installed or available')

        if (self._format == 'npy') and (np.__version__ < '1.9.0') and (self._mode in 'wa'):
            # writing to npy takes an unecessary copy in memory [1], which was fixed in version 1.9.0
            # [1] https://www.youtube.com/watch?v=TZdqeEd7iTM
            pass

    def _save_image(self, img, frame_number, frame_time):
        dest = os.path.join(self._chunk_cdir, '%06d.%s' % (self._frame_n % self._chunksize, self._format))

        if self._format in self._cv2_fmts:
            if self._format == 'ppm':
                img = ensure_color(img)
            elif self._format == 'pgm':
                img = ensure_grayscale(img)
            cv2.imwrite(dest, img)
        elif self._format == 'npy':
            np.save(dest, img)
        elif self._format == 'bpk':
            bloscpack.pack_ndarray_file(img, dest)

        self._save_image_metadata(frame_number, frame_time)

    def _save_chunk(self, old, new):
        if old is not None:
            self._save_chunk_metadata(os.path.join(self._chunk_cdir, 'index'))
        if new is not None:
            self._chunk_cdir = os.path.join(self._basedir, '%06d' % new)
            os.mkdir(self._chunk_cdir)
            self._new_chunk_metadata(os.path.join(self._basedir, '%06d' % new, 'index'))

    def _find_chunks(self, chunk_numbers):
        if chunk_numbers is None:
            immediate_dirs = next(os.walk(self._basedir))[1]
            chunk_numbers = list(map(int, immediate_dirs))  # N.B. need list, as we iterate it twice
        return list(zip(chunk_numbers, tuple(os.path.join(self._basedir, '%06d' % int(n), 'index')
                                             for n in chunk_numbers)))

    # noinspection PyShadowingBuiltins
    @staticmethod
    def _open_image(path, format, color):
        if format in DirectoryImgStore._cv2_fmts:
            flags = cv2.IMREAD_COLOR if color else cv2.IMREAD_GRAYSCALE
            img = cv2.imread(path, flags)
        elif format == 'npy':
            img = np.load(path)
        elif format == 'bpk':
            with open(path, 'rb') as reader:
                img = bloscpack.numpy_io.unpack_ndarray(bloscpack.file_io.CompressedFPSource(reader))
        else:
            # Won't get here unless we relax checks in constructor, but better safe
            raise ValueError('unknown format %s' % format)
        return img

    def _load_image(self, idx):
        path = os.path.join(self._chunk_cdir, '%06d.%s' % (idx, self._format))
        img = self._open_image(path, self._format, self._color)
        return img, (self._chunk_md['frame_number'][idx], self._chunk_md['frame_time'][idx])

    def _load_chunk(self, n):
        cdir = os.path.join(self._basedir, '%06d' % n)
        if cdir != self._chunk_cdir:
            self._log.debug('loading chunk %s' % n)
            self._chunk_cdir = cdir
            self._chunk_md = self._index.get_chunk_metadata(n)

        self._chunk_n = n
        self._chunk_current_frame_idx = -1  # not used in DirectoryImgStore, but maintain compat

    @staticmethod
    def _extract_only_frame(basedir, chunk_n, frame_n, smd):
        fmt = smd['format']
        cdir = os.path.join(basedir, '%06d' % chunk_n)
        path = os.path.join(cdir, '%06d.%s' % (frame_n, fmt))

        imgshape = tuple(smd['imgshape'])
        color = (imgshape[-1] == 3) & (len(imgshape) == 3)

        return DirectoryImgStore._open_image(path, fmt, color)

    @classmethod
    def supported_formats(cls):
        fmts = list(cls._cv2_fmts) + list(cls._raw_fmts)
        if bloscpack is None:
            fmts.remove('bpk')
        return fmts

    @property
    def lossless(self):
        return self._format != 'jpg'


class VideoImgStore(_ImgStore):
    _supported_modes = 'wr'

    _cv2_fmts = {'mjpeg': FourCC('M', 'J', 'P', 'G'),
                 'mjpeg/avi': FourCC('M', 'J', 'P', 'G'),
                 'h264/mkv': FourCC('H', '2', '6', '4'),
                 'avc1/mp4': FourCC('a', 'v', 'c', '1')}

    _DEFAULT_CHUNKSIZE = 10000

    def __init__(self, **kwargs):

        self._cap = None
        self._capfn = None

        fmt = kwargs.get('format')
        # backwards compat
        if fmt == 'mjpeg':
            kwargs['format'] = fmt = 'mjpeg/avi'

        # default to seeking enable
        seek = kwargs.pop('seek', True)
        # keep compat with VideoImgStoreFFMPEG
        kwargs.pop('gpu_id', None)

        if kwargs['mode'] == 'w':
            imgshape = kwargs['imgshape']

            if 'chunksize' not in kwargs:
                kwargs['chunksize'] = self._DEFAULT_CHUNKSIZE

            try:
                self._codec = self._cv2_fmts[fmt]
            except KeyError:
                raise ValueError('only %r supported', (self._cv2_fmts.keys(),))

            self._color = (imgshape[-1] == 3) & (len(imgshape) == 3)

            metadata = kwargs.get('metadata', {})
            metadata[STORE_MD_KEY] = {'extension': '.%s' % fmt.split('/')[1]}
            kwargs['metadata'] = metadata
            kwargs['encoding'] = kwargs.pop('encoding', None)

        _ImgStore.__init__(self, **kwargs)

        self._supports_seeking = seek
        if self._supports_seeking:
            self._log.info('seeking enabled on store')
        else:
            self._log.info('seeking NOT enabled on store (will fallback to sequential reading)')

        if self._mode == 'r':
            if self.supports_format(self._format) or \
                    ((self._metadata.get('class') == 'VideoImgStoreFFMPEG') and
                     (('264' in self._format) or ('nvenc-' in self._format))):
                check_imgshape = self._calculate_written_image_shape(self._imgshape, '')
            else:
                check_imgshape = tuple(self._imgshape)

            if check_imgshape != self._imgshape:
                self._log.warn('previous store had incorrect image_shape: corrected %r -> %r' % (
                    self._imgshape, check_imgshape))
                self._imgshape = check_imgshape
            self._color = (self._imgshape[-1] == 3) & (len(self._imgshape) == 3)

        self._log.info("store is native color: %s (or grayscale with encoding: '%s')" % (self._color, self._encoding))

    def _readability_check(self, smd_class, smd_version):
        can_read = {'VideoImgStoreFFMPEG', 'VideoImgStore',
                    getattr(self, 'class_name', self.__class__.__name__)}
        if smd_class not in can_read:
            raise ValueError('incompatible store, can_read:%r opened:%r' % (can_read, smd_class))
        if smd_version != self._version:
            raise ValueError('incompatible store version')
        return True

    def _calculate_written_image_shape(self, imgshape, fmt):
        _imgshape = list(imgshape)
        # bitwise and with -2 truncates downwards to even
        _imgshape[0] = int(_imgshape[0]) & -2
        _imgshape[1] = int(_imgshape[1]) & -2
        return tuple(_imgshape)

    @staticmethod
    def _get_chunk_extension(metadata):
        # forward compatibility
        try:
            return metadata['extension']
        except KeyError:
            # backward compatibility with old mjpeg stores
            if metadata['format'] == 'mjpeg':
                return '.avi'
            # backwards compatibility with old bview/motif stores
            return '.mp4'

    @property
    def _ext(self):
        return self._get_chunk_extension(self._metadata)

    @property
    def _chunk_paths(self):
        ext = self._ext
        return ['%s%s' % (p[1], ext) for p in self._chunk_n_and_chunk_paths]

    def _find_chunks(self, chunk_numbers):
        if chunk_numbers is None:
            avis = map(os.path.basename, glob.glob(os.path.join(self._basedir, '*%s' % self._ext)))
            chunk_numbers = list(map(int, map(operator.itemgetter(0), map(os.path.splitext, avis))))
        return list(zip(chunk_numbers, tuple(os.path.join(self._basedir, '%06d' % n) for n in chunk_numbers)))

    def _save_image(self, img, frame_number, frame_time):
        # we always write color because its more supported
        frame = ensure_color(img)
        self._cap.write(frame)
        if not os.path.isfile(self._capfn):
            raise Exception('Your opencv build does support writing this codec')
        self._save_image_metadata(frame_number, frame_time)

    def _save_chunk(self, old, new):
        if self._cap is not None:
            self._cap.release()
            self._save_chunk_metadata(os.path.join(self._basedir, '%06d' % old))

        if new is not None:
            fn = os.path.join(self._basedir, '%06d%s' % (new, self._ext))
            h, w = self._imgshape[:2]
            try:
                self._cap = cv2.VideoWriter(filename=fn,
                                            apiPreference=cv2.CAP_FFMPEG,
                                            fourcc=self._codec,
                                            fps=25,
                                            frameSize=(w, h),
                                            isColor=True)
            except TypeError:
                self._log.error('old (< 3.2) cv2 not supported (this is %r)' % (cv2.__version__,))
                self._cap = cv2.VideoWriter(filename=fn,
                                            fourcc=self._codec,
                                            fps=25,
                                            frameSize=(w, h),
                                            isColor=True)

            self._capfn = fn
            self._new_chunk_metadata(os.path.join(self._basedir, '%06d' % new))

    def _load_image(self, idx):
        if self._supports_seeking:
            # only seek if we have to, otherwise take the fast path
            if (idx - self._chunk_current_frame_idx) != 1:
                self._cap.set(getattr(cv2, "CAP_PROP_POS_FRAMES", 1), idx)
        else:
            if idx <= self._chunk_current_frame_idx:
                self._load_chunk(self._chunk_n, _force=True)

            i = self._chunk_current_frame_idx + 1
            while i < idx:
                _, img = self._cap.read()
                i += 1

        _, _img = self._cap.read()
        if self._color:
            # almost certainly no-op as opencv usually returns color frames....
            img = ensure_color(_img)
        else:
            img = ensure_grayscale(_img)

        return img, (self._chunk_md['frame_number'][idx], self._chunk_md['frame_time'][idx])

    def _load_chunk(self, n, _force=False):
        fn = os.path.join(self._basedir, '%06d%s' % (n, self._ext))
        if _force or (fn != self._capfn):
            if self._cap is not None:
                self._cap.release()

            self._log.debug('loading chunk %s' % n)
            self._capfn = fn
            # noinspection PyArgumentList
            self._cap = cv2.VideoCapture(self._capfn)
            self._chunk_current_frame_idx = -1

            if not self._cap.isOpened():
                raise Exception("OpenCV unable to open %s" % fn)

            self._chunk_md = self._index.get_chunk_metadata(n)

        self._chunk_n = n

    @classmethod
    def supported_formats(cls):
        # remove the duplicate
        fmts = list(cls._cv2_fmts.keys())
        fmts.remove('mjpeg')
        return fmts

    @classmethod
    def supports_format(cls, fmt):
        return fmt in cls._cv2_fmts

    @staticmethod
    def _extract_only_frame(basedir, chunk_n, frame_n, smd):
        capfn = os.path.join(basedir, '%06d%s' % (chunk_n,
                                                  VideoImgStore._get_chunk_extension(smd)))
        # noinspection PyArgumentList
        cap = cv2.VideoCapture(capfn)

        if _VERBOSE_DEBUG_CHUNKS:
            log = logging.getLogger('imgstore')
            log.debug('opening %s chunk %d frame_idx %d' % (capfn, chunk_n, frame_n))

        try:
            if frame_n > 0:
                cap.set(getattr(cv2, "CAP_PROP_POS_FRAMES", 1), frame_n)
            _, img = cap.read()
            return img
        finally:
            cap.release()

    @property
    def lossless(self):
        return False

    def close(self, **kwargs):
        super(VideoImgStore, self).close(**kwargs)
        if self._cap is not None:
            self._cap.release()
            self._cap = None
            self._capfn = None

    def empty(self):
        _ImgStore.empty(self)
        # use find_chunks because it removes also invalid chunks
        for _, chunk_path in self._find_chunks(chunk_numbers=None):
            os.unlink(chunk_path + self._ext)
            self._remove_index(chunk_path)

    def insert_chunk(self, video_path, frame_numbers, frame_times, move=True):
        assert len(frame_numbers) == len(frame_times)
        assert video_path.endswith(self._ext)

        self._new_chunk_metadata(os.path.join(self._basedir, '%06d' % self._chunk_n))
        self._chunk_md['frame_number'] = np.asarray(frame_numbers)
        self._chunk_md['frame_time'] = np.asarray(frame_times)
        self._save_chunk_metadata(os.path.join(self._basedir, '%06d' % self._chunk_n))

        vid = os.path.join(self._basedir, '%06d%s' % (self._chunk_n, self._ext))
        if move:
            shutil.move(video_path, vid)
        else:
            shutil.copy(video_path, vid)

        self._chunk_n += 1


def _parse_basedir_fullpath(path, basedir, read):
    # the API is not so clean that we support both path and basedir, but we have to keep
    # backwards compatibility as best as possible

    if path and basedir:
        raise ValueError('path and basedir are mutually exclusive')

    if basedir:
        if basedir.endswith(STORE_MD_FILENAME) and (os.path.basename(basedir) == STORE_MD_FILENAME):
            raise ValueError('basedir should be a directory (not a file ending with %s)' % STORE_MD_FILENAME)
        return basedir, os.path.join(basedir, STORE_MD_FILENAME)

    if path and path.endswith(STORE_MD_FILENAME) and (os.path.basename(path) == STORE_MD_FILENAME):
        basedir = os.path.dirname(path)
        fullpath = path
    elif path:
        if read:
            if os.path.isdir(path) and os.path.exists(os.path.join(path, STORE_MD_FILENAME)):
                basedir = path
                fullpath = os.path.join(path, STORE_MD_FILENAME)
            else:
                raise ValueError("path '%s' does not exist" % path)
        else:
            # does not end with STORE_MD_FILENAME
            basedir = path
            fullpath = os.path.join(path, STORE_MD_FILENAME)
    else:
        raise ValueError('should be a path to a store %s file or a directory containing one' % STORE_MD_FILENAME)

    return basedir, fullpath


def new_for_filename(path, **kwargs):
    if 'mode' not in kwargs:
        kwargs['mode'] = 'r'

    basedir, fullpath = _parse_basedir_fullpath(path, kwargs.pop('basedir', None),
                                                read=kwargs.get('mode') == 'r')

    kwargs['basedir'] = basedir

    with open(fullpath, 'rt') as f:
        clsname = yaml.load(f, Loader=yaml.SafeLoader)[STORE_MD_KEY]['class']

    # retain compatibility with internal loopbio stores
    if clsname == 'VideoImgStoreFFMPEG':
        clsname = 'VideoImgStore'

    try:
        cls = {DirectoryImgStore.__name__: DirectoryImgStore,
               VideoImgStore.__name__: VideoImgStore}[clsname]
    except KeyError:
        raise ValueError('store class %s not supported' % clsname)

    return cls(**kwargs)


def new_for_format(fmt, path=None, **kwargs):
    if 'mode' not in kwargs:
        kwargs['mode'] = 'w'

    if kwargs.get('mode') == 'r':
        return new_for_filename(path, **kwargs)

    # we are writing mode
    basedir, _ = _parse_basedir_fullpath(path, kwargs.pop('basedir', None),
                                         read=kwargs.get('mode') == 'r')
    kwargs['basedir'] = basedir

    for cls in (DirectoryImgStore, VideoImgStore):
        if cls.supports_format(fmt):
            kwargs['format'] = fmt
            return cls(**kwargs)
    raise ValueError("store class not found which supports format '%s'" % fmt)


def extract_only_frame(path, frame_index):

    _, fullpath = _parse_basedir_fullpath(path, None, True)

    smd = _extract_store_metadata(fullpath)
    clsname = smd['class']

    if clsname == 'VideoImgStoreFFMPEG':
        clsname = 'VideoImgStore'

    try:
        cls = {DirectoryImgStore.__name__: DirectoryImgStore,
               VideoImgStore.__name__: VideoImgStore}[clsname]
    except KeyError:
        raise ValueError('store class %s not supported' % clsname)

    return cls.extract_only_frame(full_path=fullpath, frame_index=frame_index, _smd=smd)


def get_supported_formats():
    f = []
    for cls in (DirectoryImgStore, VideoImgStore):
        f.extend(cls.supported_formats())
    return f



import numpy as np

import cv2

import imgstore.util as ibu


def test_codec():
    c = ibu.ImageCodecProcessor()
    assert c.cv2_enum_to_code(cv2.COLOR_RGB2BGR) == 'cv_rgb'
    assert not c.check_code('foo')
    assert c.check_code('cv_bayerbg')

    bw = np.zeros((100, 100), np.uint8)
    color = c.convert_to_bgr(bw, 'cv_bayerbg')
    assert (color.ndim == 3) and (color.shape[2] == 3) and \
           (color.shape[0] == bw.shape[0]) and (color.shape[1] == bw.shape[1])

    c = ibu.ImageCodecProcessor.from_cv2_enum(cv2.COLOR_BayerRG2BGR)
    assert c.encoding == 'cv_bayerrg'

    c = ibu.ImageCodecProcessor.from_pylon_format('BayerBG')
    assert c.encoding == 'cv_bayerrg'



import logging

import cv2
import numpy as np

from .stores import new_for_filename, get_supported_formats, new_for_format
from .ui import new_window
from .util import motif_get_parse_true_fps, motif_get_recording_timezone
from .font import FontRendererStatusBar

_log = logging.getLogger('imgstore.apps')


def main_viewer():
    import argparse
    import logging
    import datetime

    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument('path', nargs=1)
    parser.add_argument('--fps', default=30., type=float,
                        help='playback store at this speed (0 = play back as fast as possible)')
    parser.add_argument('--hide-statusbar', default=False, action='store_true',
                        help='hide status bar which shows frame time/number')
    args = parser.parse_args()

    if args.fps == 0.0:
        args.fps = 1000.0  # ensure w sleep at least 1ms

    store = new_for_filename(args.path[0])
    tz = motif_get_recording_timezone(store)

    def _draw_statusbar_nop(_img, _md):
        pass

    draw_statusbar = _draw_statusbar_nop

    if not args.hide_statusbar:
        _rend = None
        try:
            _ltxt_fmt = '{:%d}' % (len('%d' % int(store.frame_max)))
            def _get_status_bar_txt(_frame_number, _frame_time):
                return _ltxt_fmt.format(_frame_number), store.convert_frame_time_to_datetime(_frame_time, tz=tz).isoformat()

            _ltxt, _rtxt = _get_status_bar_txt(store.frame_max, store.frame_time_max)
            _rend = FontRendererStatusBar.new_for_image_size((store.image_shape[1], store.image_shape[0]),
                                                             ltext=_ltxt, rtext=_rtxt)
        except Exception:
            _log.error('could not construct statusbar', exc_info=True)

        def _draw_statusbar_actual(_img, _md):
            if _rend is not None:
                _ltxt, _rtxt = _get_status_bar_txt(*md)
                return _rend.render(img=_img, ltext=_ltxt, rtext=_rtxt)
            return _img

        draw_statusbar = _draw_statusbar_actual

    _log.info('true fps: %s' % motif_get_parse_true_fps(store))
    _log.info('fr fps: %s' % (1.0 / np.median(np.diff(store._get_chunk_metadata(0)['frame_time']))))

    win = new_window('imgstore', shape=store.image_shape)
    _log.debug('created window: %r' % win)

    paused = False
    while True:
        if not paused:
            try:
                img, md = store.get_next_image()
            except EOFError:
                break

        try:
            draw_statusbar(img, md)
        except:
            pass

        win.imshow('imgstore', img)

        k = cv2.waitKey(int(1000. / args.fps)) & 0xFF
        if (k == 27) or (k == ord('q')):
            break

        if k == ord(' '):
            paused ^= 1


def main_saver():
    import os.path
    import argparse
    import logging
    import time
    import itertools
    import errno

    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument('dest', nargs=1)
    parser.add_argument('--source', type=str, default=0,
                        metavar='PATH',
                        help='path to a file to convert to an imgstore')
    parser.add_argument('--format',
                        choices=get_supported_formats(), default='mjpeg')
    args = parser.parse_args()

    # noinspection PyArgumentList
    cap = cv2.VideoCapture(args.source)
    _, img = cap.read()

    if img is None:
        parser.error('could not open source: %s' % args.source)

    path = args.dest[0]
    if os.path.exists(path):
        parser.error('destination already exists')
    else:
        try:
            os.makedirs(path)
        except OSError as exc:  # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                parser.error('could not create destination path')

    store = new_for_format(args.format,
                           basedir=path, mode='w',
                           imgdtype=img.dtype,
                           imgshape=img.shape)

    for i in itertools.count():
        _, img = cap.read()

        cv2.imshow('preview', img)

        k = cv2.waitKey(1) & 0xFF
        if k == 27:
            break

        store.add_image(img, i, time.time())

    store.close()


def main_test():
    import pytest
    import os.path

    pytest.main(['-v', os.path.join(os.path.abspath(os.path.dirname(__file__)), 'tests')])


def generate_timecodes(store, dest_file):
    _ts = np.asarray(store.get_frame_metadata()['frame_time'])
    ts = _ts - _ts[0]

    dest_file.write('# timecode format v2\n')
    for t in ts:
        dest_file.write('{0:.3f}\n'.format(t * 1000.))
    return ts


def main_generate_timecodes():
    import sys
    import argparse
    import logging

    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument('path', nargs=1)
    args = parser.parse_args()

    store = new_for_filename(args.path[0])

    generate_timecodes(store, sys.stdout)



import os.path
import sqlite3
import logging
import operator

import yaml
import numpy as np

from .constants import FRAME_MD


def _load_index(path_without_extension):
    for extension in ('.npz', '.yaml'):
        path = path_without_extension + extension
        if os.path.exists(path):
            if extension == '.yaml':
                with open(path, 'rt') as f:
                    dat = yaml.safe_load(f)
                    return {k: dat[k] for k in FRAME_MD}
            elif extension == '.npz':
                with open(path, 'rb') as f:
                    dat = np.load(f)
                    return {k: dat[k].tolist() for k in FRAME_MD}

    raise IOError('could not find index %s' % path_without_extension)


# noinspection SqlNoDataSourceInspection,SqlDialectInspection,SqlResolve
class ImgStoreIndex(object):

    VERSION = '1'

    log = logging.getLogger('imgstore.index')

    def __init__(self, db=None):
        self._conn = db

        cur = self._conn.cursor()
        cur.execute('pragma query_only = ON;')

        cur.execute('SELECT value FROM index_information WHERE name = ?', ('version', ))
        v, = cur.fetchone()
        if v != self.VERSION:
            raise IOError('incorrect index version: %s vs %s' % (v, self.VERSION))

        cur.execute('SELECT COUNT(1) FROM frames')
        self.frame_count, = cur.fetchone()

        def _summary(_what):
            cur.execute('SELECT value FROM summary WHERE name = ?', (_what,))
            return cur.fetchone()[0]

        if self.frame_count:
            self.frame_time_max = _summary('frame_time_max')
            self.frame_time_min = _summary('frame_time_min')
            self.frame_max = _summary('frame_max')
            self.frame_min = _summary('frame_min')

            # keep back compat for nan as types (inf -> nan)
            if not np.isreal(self.frame_max):
                self.frame_max = np.nan
            if not np.isreal(self.frame_min):
                self.frame_min = np.nan
        else:
            self.frame_max = self.frame_min = np.nan
            self.frame_time_max = self.frame_time_min = 0.0

        self.log.debug('frame range %f -> %f' % (self.frame_min, self.frame_max))

        # # all chunks in the store [0,1,2, ... ]
        cur.execute('SELECT chunk FROM chunks ORDER BY chunk;')
        self._chunks = tuple(row[0] for row in cur)

    @classmethod
    def create_database(cls, conn):
        c = conn.cursor()
        # Create tables
        c.execute('CREATE TABLE frames '
                  '(chunk INTEGER, frame_idx INTEGER, frame_number INTEGER, frame_time REAL)')
        c.execute('CREATE TABLE chunks '
                  '(chunk INTEGER, chunk_path TEXT)')
        c.execute('CREATE TABLE index_information '
                  '(name TEXT, value TEXT)')
        c.execute('CREATE TABLE summary '
                  '(name TEXT, value REAL)')
        c.execute('INSERT into index_information VALUES (?, ?)', ('version', cls.VERSION))
        c.execute("CREATE INDEX chunk_index ON frames (chunk, frame_idx);")
        conn.commit()

    @classmethod
    def new_from_chunks(cls, chunk_n_and_chunk_paths):
        db = sqlite3.connect(':memory:', check_same_thread=False)
        cls.create_database(db)

        frame_count = 0
        frame_max = -np.inf
        frame_min = np.inf
        frame_time_max = -np.inf
        frame_time_min = np.inf

        cur = db.cursor()

        for chunk_n, chunk_path in sorted(chunk_n_and_chunk_paths, key=operator.itemgetter(0)):
            try:
                idx = _load_index(chunk_path)
            except IOError:
                cls.log.warn('missing index for chunk %s' % chunk_n)
                continue

            if not idx['frame_number']:
                # empty chunk
                continue

            frame_count += len(idx['frame_number'])
            frame_time_min = min(frame_time_min, np.min(idx['frame_time']))
            frame_time_max = max(frame_time_max, np.max(idx['frame_time']))
            frame_min = min(frame_min, np.min(idx['frame_number']))
            frame_max = max(frame_max, np.max(idx['frame_number']))

            try:
                records = [(chunk_n, i, fn, ft) for i, (fn, ft) in enumerate(zip(idx['frame_number'],
                                                                                 idx['frame_time']))]
            except TypeError:
                cls.log.error('corrupt chunk', exc_info=True)
                continue

            cur.executemany('INSERT INTO frames VALUES (?,?,?,?)', records)
            cur.execute('INSERT INTO chunks VALUES (?, ?)', (chunk_n, chunk_path))

            db.commit()

        cur.execute('INSERT INTO summary VALUES (?,?)', ('frame_time_min', float(frame_time_min)))
        cur.execute('INSERT INTO summary VALUES (?,?)', ('frame_time_max', float(frame_time_max)))
        cur.execute('INSERT INTO summary VALUES (?,?)', ('frame_min', float(frame_min)))
        cur.execute('INSERT INTO summary VALUES (?,?)', ('frame_max', float(frame_max)))

        db.commit()

        return cls(db)

    @classmethod
    def new_from_file(cls, path):
        db = sqlite3.connect(path, check_same_thread=False)
        return cls(db)

    @staticmethod
    def _get_metadata(cur):
        md = {'frame_number': [], 'frame_time': []}
        for row in cur:
            md['frame_number'].append(row[0])
            md['frame_time'].append(row[1])
        return md

    @property
    def chunks(self):
        """ the number of non-empty chunks that contain images """
        return self._chunks

    def to_file(self, path):
        db = sqlite3.connect(path)
        with db:
            for line in self._conn.iterdump():
                # let python handle the transactions
                if line not in ('BEGIN;', 'COMMIT;'):
                    db.execute(line)
        db.commit()
        db.close()

    def get_all_metadata(self):
        cur = self._conn.cursor()
        cur.execute("SELECT frame_number, frame_time FROM frames ORDER BY rowid;")
        return self._get_metadata(cur)

    def get_chunk_metadata(self, chunk_n):
        cur = self._conn.cursor()
        cur.execute("SELECT frame_number, frame_time FROM frames WHERE chunk = ? ORDER BY rowid;", (chunk_n, ))
        return self._get_metadata(cur)

    def find_chunk(self, what, value):
        assert what in ('frame_number', 'frame_time', 'index')
        cur = self._conn.cursor()

        if what == 'index':
            cur.execute("SELECT chunk, frame_idx FROM frames ORDER BY rowid LIMIT 1 OFFSET {};".format(int(value)))
        else:
            cur.execute("SELECT chunk, frame_idx FROM frames WHERE {} = ?;".format(what), (value, ))

        try:
            chunk_n, frame_idx = cur.fetchone()
        except TypeError:  # no result
            return -1, -1

        return chunk_n, frame_idx

    def find_chunk_nearest(self, what, value):
        assert what in ('frame_number', 'frame_time')
        cur = self._conn.cursor()
        cur.execute("SELECT chunk, frame_idx FROM frames ORDER BY ABS(? - {}) LIMIT 1;".format(what), (value, ))
        chunk_n, frame_idx = cur.fetchone()
        return chunk_n, frame_idx


from .stores import new_for_filename, new_for_format, extract_only_frame, get_supported_formats,\
    VideoImgStore, DirectoryImgStore
from .apps import main_test
from .util import ensure_color, ensure_grayscale

test = main_test

# use only single quotes here (you can write beta versions like -beta4 etc)
__version__ = '0.4.0'



from subprocess import DEVNULL

STORE_MD_KEY = '__store'
STORE_MD_FILENAME = 'metadata.yaml'
STORE_LOCK_FILENAME = '.lock'
STORE_INDEX_FILENAME = '.index.sqlite'

EXTRA_DATA_FILE_EXTENSIONS = ('.extra.json', '.extra_data.json', '.extra_data.h5')

FRAME_MD = ('frame_number', 'frame_time')




import re
import sys
import logging
import subprocess

import cv2
import numpy as np

_SCREEN_RESOLUTION = None
_IS_MAC = sys.platform == 'darwin'

_log = logging.getLogger('imgstore.ui')


class _Window(object):
    def __init__(self, name, flags, sizestr):
        self.name = name
        self._flags = flags
        self._size = sizestr
        self._set = False

    def __repr__(self):
        return "Window<'%s', flags=%s, sizestr=%s, mac=%s>" % (self.name, bin(self._flags),
                                                               self._size, _IS_MAC)

    def __str__(self):
        return self.name

    @property
    def size(self):
        if 'x' in self._size:
            return tuple(map(int, self._size.split('x')))
        return None

    def imshow(self, *args, **kwargs):
        cv2.imshow(*args, **kwargs)
        # Mac requires window size to be set after the first imshow
        if _IS_MAC and not self._set:
            sz = self.size
            if sz is not None:
                w, h = sz
                cv2.resizeWindow(self.name, int(w), int(h))
                self._set = True

    # noinspection PyPep8Naming
    def waitKey(self, *args, **kwargs):
        return cv2.waitKey(*args, **kwargs)


def get_screen_resolution():
    """
    Returns: the screen resolution in pixels, as a string WxH
    """
    global _SCREEN_RESOLUTION

    if _SCREEN_RESOLUTION is None:
        resolution = ''

        try:
            if _IS_MAC:
                _re = re.compile(r"""^.*Resolution:\s*([\d\sx]+)""")
                out = subprocess.check_output(['system_profiler', 'SPDisplaysDataType'])
                for l in out.splitlines():
                    _s = l.decode('utf-8')
                    m = _re.match(_s)
                    if m:
                        resolution = m.groups()[0]
            else:
                out = subprocess.check_output(['xrandr'], text=True)
                resolution_line = [l for l in out.splitlines() if '*' in l][0]
                resolution = resolution_line.split()[0]
        except subprocess.CalledProcessError as exc:
            _log.warn("could not detect resolution: %r:" % exc)
        except Exception as exc:
            _log.warn("could not parse resolution: %r:" % exc)

        _SCREEN_RESOLUTION = resolution

    return _SCREEN_RESOLUTION


def get_and_parse_screen_resolution(scale=1.0, default=(1024, 768)):
    res = get_screen_resolution()
    _w, _h = default

    if res:
        try:
            _w, _h = map(float, res.split('x'))
        except ValueError:
            _log.warn("could not splitting resolution string: %r:" % res)
    else:
        _log.warn("could not get resolution string")

    w = float(scale) * _w
    h = float(scale) * _h
    return int(w), int(h)


def new_window(name, size=None, expanded_ui=True, shape=None, screen_relative_size=0.75):
    if shape is not None:
        size = shape[1], shape[0]

    flags = cv2.WINDOW_GUI_EXPANDED if expanded_ui else cv2.WINDOW_GUI_NORMAL
    if size is not None:
        flags = cv2.WINDOW_NORMAL | flags
        cv2.namedWindow(name, flags)

        if (size[0] > 0) and (not np.isnan(size[0])):

            # create a resizable window but limit its size to 75% the screen size
            sw, sh = get_and_parse_screen_resolution(scale=screen_relative_size, default=(1024, 768))

            if np.isinf(size[0]):
                w = sw
                h = sh
            else:
                w, h = size

                sfw = (float(sw) / w) if w > sw else 1.0
                sfh = (float(sh) / h) if h > sh else 1.0

                sf = min(sfw, sfh)

                h *= sf
                w *= sf

            cv2.resizeWindow(name, int(w), int(h))
            sizestr = '%dx%d' % (int(w), int(h))
        else:
            sizestr = 'unknown'
    else:
        flags = cv2.WINDOW_AUTOSIZE | flags
        cv2.namedWindow(name, flags)
        sizestr = 'auto'

    win = _Window(name, flags, sizestr)
    return win



import numpy as np
import numpy.testing as npt
import os.path
import shutil
import tempfile
import time
import itertools

import cv2
import pytest

from imgstore import stores
from imgstore.util import FourCC, ensure_color, ensure_grayscale
from imgstore.tests import TEST_DATA_DIR


def encode_image(num, nbits=16, imgsize=512):
    # makes a square image that looks a bit like a barcode. colums of the matrix all 255
    # if the bit in the number is 1, 0 otherwise
    row = np.fromiter((255*int(c) for c in ('{0:0%db}' % nbits).format(num)), dtype=np.uint8)
    mat = np.tile(row, (nbits, 1))
    return cv2.resize(mat, dsize=(imgsize, imgsize), interpolation=cv2.INTER_NEAREST)


def decode_image(img, nbits=16, imgsize=512):
    h, w = img.shape[:2]
    assert (h == imgsize) and (w == imgsize)
    assert len(img.shape) == 2

    img = cv2.resize(img, dsize=(nbits, nbits), interpolation=cv2.INTER_NEAREST)
    row = (np.mean(img, axis=0) > 127).astype(np.uint8)
    bstr = ''.join(str(v) for v in row)
    return int(bstr, 2)


def new_framecode_video(dest, frame0, nframes):
    assert dest.endswith('.avi')
    cap = cv2.VideoWriter(dest, FourCC('M', 'J', 'P', 'G'), 25, (512, 512), isColor=True)
    for i in range(frame0, frame0+nframes):
        cap.write(ensure_color(encode_image(i, imgsize=512)))
    cap.release()
    return dest


def get_size(start_path):
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    return total_size


def new_framecode_store(dest, frame0, nframes, format='png', chunksize=7):
    kwargs = dict(basedir=dest,
                  mode='w',
                  imgshape=(512, 512, 3),
                  imgdtype=np.uint8,
                  chunksize=chunksize,
                  format=format)

    d = stores.new_for_format(format, **kwargs)
    for i in range(frame0, frame0+nframes):
        frame = ensure_color(encode_image(i, imgsize=512))
        d.add_image(frame, i, time.time())
    d.close()

    return d, dest


@pytest.fixture
def loglevel_info():
    import logging
    logging.basicConfig(level=logging.INFO)


@pytest.fixture
def loglevel_debug():
    import logging
    logging.basicConfig(level=logging.DEBUG)


@pytest.fixture
def grey_image():
    cimg = cv2.imread(os.path.join(TEST_DATA_DIR, 'graffiti.png'), cv2.IMREAD_GRAYSCALE)
    return cv2.resize(cimg, (1920, 1200))


@pytest.fixture
def graffiti():
    return cv2.imread(os.path.join(TEST_DATA_DIR, 'graffiti.png'), cv2.IMREAD_COLOR)


@pytest.mark.parametrize('fmt', stores.get_supported_formats())
def test_all(tmpdir, fmt):
    F = 21
    SZ = 512
    imgtype = 'color'

    tdir = tmpdir.strpath

    def _build_img(num):
        img = encode_image(num, imgsize=SZ)
        if imgtype == 'color':
            return np.dstack((img, img, img))
        return img

    def _decode_image(img):
        if imgtype == 'color':
            return decode_image(img[:, :, 0], imgsize=SZ)
        return decode_image(img, imgsize=SZ)

    orig_img = _build_img(0)
    kwargs = dict(basedir=tdir,
                  mode='w',
                  imgshape=orig_img.shape,
                  imgdtype=orig_img.dtype,
                  chunksize=10,
                  metadata={'timezone': 'Europe/Austria'},
                  format=fmt)
    d = stores.new_for_format(fmt, **kwargs)

    fns = []
    frame_times = {}
    for i in range(F):
        t = time.time()
        d.add_image(_build_img(i), i, t)
        frame_times[i] = t
        fns.append(i)
    d.close()

    d = stores.new_for_filename(os.path.join(d.filename, stores.STORE_MD_FILENAME), mode='r')
    assert not d.has_extra_data

    assert d.user_metadata['timezone'] == 'Europe/Austria'

    r = np.random.RandomState(42)
    r.shuffle(fns)
    for f in fns:
        img, (frame_number, frame_time) = d.get_image(frame_number=f)
        assert f == frame_number
        assert frame_times[f] == frame_time
        assert img.shape == orig_img.shape
        assert _decode_image(img) == f

    d.close()


@pytest.mark.parametrize('fmt', ('mjpeg', 'npy', 'h264/mkv', 'avc1/mp4'))
@pytest.mark.parametrize('imgtype', ('b&w', 'color'))
def test_outoforder(tmpdir,  fmt, imgtype):
    SZ = 512

    tdir = tmpdir.strpath

    def _build_img(num):
        img = encode_image(num, imgsize=SZ)
        if imgtype == 'color':
            return np.dstack((img, img, img))
        return img

    def _decode_image(img):
        if imgtype == 'color':
            return decode_image(img[:, :, 0], imgsize=SZ)
        return decode_image(img, imgsize=SZ)

    orig_img = _build_img(0)
    kwargs = dict(basedir=tdir,
                  mode='w',
                  imgshape=orig_img.shape,
                  imgdtype=orig_img.dtype,
                  chunksize=5,
                  format=fmt)

    d = stores.new_for_format(fmt, **kwargs)

    assert d.image_shape == orig_img.shape
    assert os.path.isfile(os.path.join(d.filename, stores.STORE_LOCK_FILENAME))

    F = 50
    assert F > 20+1
    assert F < 255

    fns = []
    frame_times = {}
    frame_extra = {}
    for i in range(5):
        t = time.time()

        # store the actual frame number
        img = _build_img(i)

        d.add_image(img, i, t)
        fns.append(i)
        frame_times[i] = t

        edat = dict(foo=123, N=i)
        d.add_extra_data(**edat)
        frame_extra[i] = edat

    for i in (6, 8, 9, 11):
        t = time.time()
        img = _build_img(i)
        d.add_image(img, i, t)
        fns.append(i)
        frame_times[i] = t

        edat = dict(foo=123, N=i)
        d.add_extra_data(**edat)
        frame_extra[i] = edat

    for i in range(20, 15, -1):
        t = time.time()
        img = _build_img(i)
        d.add_image(img, i, t)
        fns.append(i)
        frame_times[i] = t
    for i in range(20+1, F+1):
        t = time.time()
        img = _build_img(i)
        d.add_image(img, i, t)
        fns.append(i)
        frame_times[i] = t

    d.close()
    assert not os.path.isfile(os.path.join(d.filename, stores.STORE_LOCK_FILENAME))

    d = stores.new_for_filename(os.path.join(d.filename, stores.STORE_MD_FILENAME))

    # read mode doesnt create lock file
    assert not os.path.isfile(os.path.join(d.filename, stores.STORE_LOCK_FILENAME))

    assert d.image_shape == orig_img.shape
    assert d.has_extra_data
    assert d.frame_min == 0
    assert d.frame_max == F

    r = np.random.RandomState(42)
    r.shuffle(fns)
    for f in fns:
        img, (frame_number, frame_time) = d.get_image(frame_number=f)
        assert d.frame_number == f
        assert f == frame_number
        assert frame_times[f] == frame_time
        assert img.shape == orig_img.shape
        assert _decode_image(img) == f

    md = d.get_frame_metadata()
    npt.assert_array_equal(sorted([frame_times[f] for f in fns]), sorted(md['frame_time']))
    npt.assert_array_equal(sorted(fns), sorted(md['frame_number']))

    df = d.get_extra_data()
    assert len(df) == len(frame_extra)

    df.set_index('frame_number', inplace=True, verify_integrity=True)
    for _i, _v in frame_extra.items():
        _row = df.loc[_i]
        assert _row['foo'] == _v['foo']
        assert _row['N'] == _i

    sz_bytes = get_size(tdir)
    cmp_ratio = abs(100 - (100.0 * ((len(fns) * float(orig_img.nbytes)) / sz_bytes)))

    print("\n%s %s %dx%dpx@%s frames (size: %.1fMB, %.1f%% %scompression)" % (
        fmt, F,
        orig_img.shape[1], orig_img.shape[0], orig_img.dtype,
        sz_bytes / (1024 * 1024.),
        cmp_ratio,
        '' if d.lossless else 'LOSSY '))


@pytest.mark.xfail(strict=False)
def test_create_and_times(loglevel_debug, tmpdir):
    import pytz

    tdir = tmpdir.strpath

    # check we always get a UTC timesone
    # does not have created_utc in metadata
    d = stores.new_for_filename(os.path.join(TEST_DATA_DIR, 'store_mp4', 'metadata.yaml'))
    c_utc, tz = d.created
    assert c_utc.tzinfo is not None
    assert c_utc.tzinfo == pytz.utc

    store, _ = new_framecode_store(dest=tdir, frame0=0, nframes=1)
    assert store.uuid is not None

    store2 = stores.new_for_filename(store.full_path)
    assert store2.uuid == store.uuid
    assert store2.created == store.created

    c_utc, tz = store2.created
    assert c_utc.tzinfo is not None
    assert c_utc.tzinfo == pytz.utc


def test_testencode_decode():
    L = 16
    SZ = 512

    for i in (0, 1, 13, 127, 255, 34385, (2 ** L) - 1):
        img = encode_image(num=i, nbits=L, imgsize=SZ)
        v = decode_image(img, nbits=L, imgsize=SZ)
        assert v == i


@pytest.mark.parametrize('fmt', ('npy', 'mjpeg', 'h264/mkv', 'avc1/mp4'))
def test_extract_only(loglevel_debug, fmt, tmpdir):
    tdir = tmpdir.strpath

    d, _ = new_framecode_store(dest=tdir,
                               frame0=57, nframes=20, format=fmt)

    for i, fn in enumerate(range(57, 57+20)):
        img = stores.extract_only_frame(d.full_path, i)
        assert decode_image(ensure_color(img)[:, :, 0]) == fn


def test_extract_only_motif(loglevel_debug):
    full_path = os.path.join(TEST_DATA_DIR, 'store_mp4', 'metadata.yaml')
    img = stores.extract_only_frame(full_path, 42)
    assert decode_image(ensure_color(img)[:, :, 0]) == 42

    basedir = os.path.join(TEST_DATA_DIR, 'store_mp4')
    img = stores.extract_only_frame(basedir, 42)
    assert decode_image(ensure_color(img)[:, :, 0]) == 42


def test_manual_assembly(loglevel_debug, tmpdir):
    tdir = tmpdir.strpath

    a = new_framecode_video(dest=os.path.join(tdir, 'a.avi'),
                            frame0=0, nframes=10)

    a_fns = list(range(0, 10))  # [0 ... 9]

    b = new_framecode_video(dest=os.path.join(tdir, 'b.avi'),
                            frame0=57, nframes=20)

    b_fns = list(range(57, 57 + 20))  # [57 ... 76]

    dest = os.path.join(tdir, 'store')
    store = stores.VideoImgStore(basedir=dest,
                                 mode='w',
                                 imgshape=(512, 512, 3),
                                 imgdtype=np.uint8,
                                 chunksize=100,
                                 format='mjpeg')
    store.empty()

    store.insert_chunk(a, a_fns, 1000.0 * np.asarray(a_fns))
    store.insert_chunk(b, b_fns, 1000.0 * np.asarray(b_fns))
    store.close()

    store = stores.new_for_filename(store.full_path)

    fns = list(itertools.chain.from_iterable((a_fns, b_fns)))
    r = np.random.RandomState(42)
    r.shuffle(fns)

    for f in fns:
        img, (frame_number, frame_time) = store.get_image(frame_number=f, exact_only=True)
        assert frame_number == f
        assert frame_time == (1000.0 * f)
        assert decode_image(img[:, :, 0]) == f

    store.close()


def test_store_frame_metadata(tmpdir):
    SZ = 512

    tdir = tmpdir.strpath

    d = stores.DirectoryImgStore(basedir=tdir,
                                 mode='w',
                                 imgshape=(SZ, SZ),
                                 imgdtype=np.uint8,
                                 chunksize=5,
                                 format='npy')

    r = np.random.RandomState(42)

    N = 100

    fns = []
    frame_times = {}
    for i in range(N):

        img = encode_image(num=i, imgsize=SZ)

        # skip some frames
        if r.rand() < 0.3:  # skip 30% of frames
            continue

        frame_number = i
        frame_time = time.time()
        d.add_image(img, frame_number, frame_time)
        fns.append(frame_number)
        frame_times[i] = frame_time

    d.close()

    assert d.frame_count == len(fns)

    d = stores.DirectoryImgStore(basedir=tdir, mode='r')

    assert d.frame_count == len(fns)

    assert d.image_shape == (SZ, SZ)

    assert d.frame_min == fns[0]
    assert d.frame_max == fns[-1]

    for f in range(N):
        img, (frame_number, frame_time) = d.get_image(frame_number=f, exact_only=False)

        if f in frame_times:
            # we skipped this frame
            assert d.frame_number == f
            assert f == frame_number
            assert frame_times[f] == frame_time

        assert img.shape == (SZ, SZ)

        sfn = decode_image(img, imgsize=SZ)
        assert int(sfn) == frame_number

    d = stores.DirectoryImgStore(basedir=tdir, mode='r')

    fniter = iter(fns)
    while True:
        try:
            img, (frame_number, frame_time) = d.get_next_image()
            assert frame_number == next(fniter)
        except EOFError:
            break

    d.close()


@pytest.mark.parametrize("seek", [True, False])
def test_videoimgstore_mp4(seek):
    L = 16
    SZ = 512

    d = stores.new_for_filename(os.path.join(TEST_DATA_DIR, 'store_mp4', 'metadata.yaml'),
                                seek=seek)
    assert d.frame_max == 178
    assert d.frame_min == 0
    assert d.chunks == [0, 1]

    for i in range(3):
        img, (_frame_number, _frame_timestamp) = d.get_next_image()
        assert img.shape == (SZ, SZ)
        assert _frame_number == i
        assert decode_image(img, nbits=L, imgsize=SZ) == i

    fns = list(range(0, 178))
    r = np.random.RandomState(42)
    r.shuffle(fns)

    for i in fns:
        img, (_frame_number, _frame_timestamp) = d.get_image(i)
        assert img.shape == (SZ, SZ)
        assert _frame_number == i
        assert decode_image(img, nbits=L, imgsize=SZ) == i


@pytest.mark.parametrize('chunksize', (2, 3, 10))
def test_reindex_to_zero(loglevel_debug, tmpdir, grey_image, chunksize):
    tdir = tmpdir.strpath

    fmt = 'mjpeg'
    kwargs = dict(basedir=tdir,
                  mode='w',
                  imgshape=grey_image.shape,
                  imgdtype=grey_image.dtype,
                  chunksize=chunksize,
                  metadata={'timezone': 'Europe/Austria'},
                  format=fmt)

    d = stores.new_for_format(fmt, **kwargs)

    d.add_image(grey_image, 3, time.time())
    d.add_extra_data(ofn=3)
    d.add_image(grey_image, 4, time.time())
    d.add_extra_data(ofn=4)
    d.add_image(grey_image, 0, time.time())
    d.add_image(grey_image, 1, time.time())
    d.add_image(grey_image, 2, time.time())
    d.add_image(grey_image, 3, time.time())
    d.add_image(grey_image, 4, time.time())
    d.add_extra_data(ofn=42)
    d.close()

    s = stores.new_for_filename(d.full_path, mode='r')
    npt.assert_array_equal(s.get_frame_metadata()['frame_number'], [3, 4, 0, 1, 2, 3, 4])

    df = s.get_extra_data()
    assert len(df) == 3
    assert sorted(df['ofn'].tolist()) == [3, 4, 42]
    assert sorted(df['frame_index'].tolist()) == [0, 1, 6]
    assert sorted(df['frame_number'].tolist()) == [3, 4, 4]

    s.reindex()
    s.close()

    j = stores.new_for_filename(s.full_path, mode='r')
    npt.assert_array_equal(j.get_frame_metadata()['frame_number'], [-2, -1, 0, 1, 2, 3, 4])

    df = j.get_extra_data()
    assert len(df) == 3
    assert sorted(df['ofn'].tolist()) == [3, 4, 42]
    assert sorted(df['frame_index'].tolist()) == [0, 1, 6]
    assert sorted(df['frame_number'].tolist()) == [-2, -1, 4]


def test_reindex_impossible(tmpdir, grey_image):
    tdir = tmpdir.strpath

    fmt = 'mjpeg'
    kwargs = dict(basedir=tdir,
                  mode='w',
                  imgshape=grey_image.shape,
                  imgdtype=grey_image.dtype,
                  chunksize=3,
                  metadata={'timezone': 'Europe/Austria'},
                  format=fmt)

    d = stores.new_for_format(fmt, **kwargs)
    d.add_image(grey_image, 3, time.time())
    d.add_image(grey_image, 4, time.time())
    d.add_image(grey_image, 0, time.time())
    d.add_image(grey_image, 0, time.time())
    d.add_image(grey_image, 1, time.time())
    d.close()

    s = stores.new_for_filename(d.full_path, mode='r')
    with pytest.raises(ValueError):
        s.reindex()
    s.close()


@pytest.mark.parametrize("seek", [True, False])
@pytest.mark.parametrize("chunksize", [7, 20, 100])
@pytest.mark.parametrize("fmt", ['npy', 'mjpeg', 'h264/mkv', 'avc1/mp4'])
def test_seek_types(loglevel_debug, tmpdir, chunksize, fmt, seek):
    tdir = tmpdir.strpath

    def _decode_image(_img):
        return decode_image(_img[:, :, 0], imgsize=512)

    kwargs = dict(basedir=tdir,
                  mode='w',
                  imgshape=(512, 512, 3),
                  imgdtype=np.uint8,
                  chunksize=chunksize,
                  format=fmt)

    d = stores.new_for_format(fmt, **kwargs)

    skips = range(2030, 2070)
    for i in range(2000, 2100):
        if i in skips:
            continue
        d.add_image(ensure_color(encode_image(i, imgsize=512)), frame_number=i, frame_time=time.time())
    d.close()

    d = stores.new_for_filename(d.full_path, seek=seek)
    assert len(d) == (100 - len(skips))
    assert d.frame_min == 2000
    assert d.frame_max == 2099

    # check seeking works
    frame, (frame_number, frame_timestamp) = d.get_image(frame_number=2000, exact_only=True)
    assert frame_number == 2000
    assert _decode_image(frame) == 2000

    frame, (frame_number, frame_timestamp) = d.get_image(frame_number=2001, exact_only=True)
    assert frame_number == 2001
    assert _decode_image(frame) == 2001

    # read next frame
    frame, (frame_number, frame_timestamp) = d.get_next_image()
    assert frame_number == 2002
    assert _decode_image(frame) == 2002

    # seek missing frame
    with pytest.raises(ValueError):
        frame, (frame_number, frame_timestamp) = d.get_image(frame_number=skips[0], exact_only=True)

    # check seeking works by frame_index
    for i in range(10, 15):
        frame, (frame_number, frame_timestamp) = d.get_image(frame_number=None, exact_only=True, frame_index=i)
        assert frame_number == 2000 + i
        assert _decode_image(frame) == 2000 + i

    # check seeking works by frame index when non monotonically increasing
    frame, (frame_number, frame_timestamp) = d.get_image(frame_number=None, exact_only=True, frame_index=29)
    assert frame_number == 2029
    assert _decode_image(frame) == 2029

    frame, (frame_number, frame_timestamp) = d.get_image(frame_number=None, exact_only=True, frame_index=30)
    assert frame_number == 2070
    assert _decode_image(frame) == 2070

    frame, (frame_number, frame_timestamp) = d.get_image(frame_number=None, exact_only=True, frame_index=31)
    assert frame_number == 2071
    assert _decode_image(frame) == 2071

    frame, (frame_number, frame_timestamp) = d.get_image(frame_number=None, exact_only=True, frame_index=0)
    assert frame_number == 2000
    assert _decode_image(frame) == 2000

    frame_count = 100 - len(skips)
    assert d.frame_count == frame_count

    # last frame
    frame, (frame_number, frame_timestamp) = d.get_image(frame_number=None, exact_only=True,
                                                         frame_index=frame_count - 1)
    assert frame_number == 2099
    assert _decode_image(frame) == 2099

    with pytest.raises(EOFError):
        frame, (frame_number, frame_timestamp) = d.get_next_image()

    with pytest.raises(ValueError):
        frame, (frame_number, frame_timestamp) = d.get_image(frame_number=None, exact_only=True,
                                                             frame_index=frame_count)


def test_always_supported():
    fmts = stores.get_supported_formats()
    assert 'mjpeg/avi' in fmts
    assert 'npy' in fmts


@pytest.mark.parametrize('fmt', stores.VideoImgStore.supported_formats())
@pytest.mark.parametrize("seek", [True, False])
def test_videoseek_extensive(loglevel_debug, fmt, seek, tmpdir):
    F = 1000
    S_PCT = 0.4

    SZ = 32

    tdir = tmpdir.strpath

    def _build_img(num):
        return ensure_color(encode_image(num, imgsize=SZ))

    def _decode_image(img):
        return decode_image(img[:, :, 0], imgsize=SZ)

    orig_img = _build_img(0)
    kwargs = dict(basedir=tdir,
                  mode='w',
                  imgshape=orig_img.shape,
                  imgdtype=orig_img.dtype,
                  chunksize=F,  # one chunk to test seeking of video file
                  format=fmt)

    with stores.new_for_format(fmt, **kwargs) as d:
        for i in range(F):
            t = time.time()
            d.add_image(_build_img(i), i, t)

    r = np.random.RandomState(0)
    fns = r.randint(low=0, high=F, size=int(S_PCT * F))

    with stores.new_for_filename(tdir, seek=seek) as d:
        for f in fns:
            img, _ = d.get_image(frame_number=f)
            assert _decode_image(img) == f


def test_new_apis(tmpdir):
    p = os.path.join(TEST_DATA_DIR, 'store_mp4')
    fullp = os.path.join(p, stores.STORE_MD_FILENAME)

    # new_for_filename (read)

    # read pass both paths
    with pytest.raises(ValueError):
        stores.new_for_filename(p, basedir=p)

    # read pass no path
    with pytest.raises(ValueError):
        stores.new_for_filename(None)

    # read pass path=basedir
    d = stores.new_for_filename(p)
    assert d.full_path == fullp
    assert d.mode == 'r'

    # read pass path=fullpath
    d = stores.new_for_filename(fullp)
    assert d.full_path == fullp
    assert d.mode == 'r'

    # read pass basedir=basedir
    d = stores.new_for_filename(None, basedir=p)
    assert d.full_path == fullp
    assert d.mode == 'r'

    with pytest.raises(ValueError):
        # read pass basedir=fullpath
        d = stores.new_for_filename(None, basedir=fullp)

    # new_for_format (read)

    d = stores.new_for_format(fmt=None, path=None, mode='r', basedir=p)
    assert d.full_path == fullp
    assert d.mode == 'r'

    d = stores.new_for_format(fmt=None, path=p, mode='r')
    assert d.full_path == fullp
    assert d.mode == 'r'

    d = stores.new_for_format(fmt=None, path=fullp, mode='r')
    assert d.full_path == fullp
    assert d.mode == 'r'

    # == new_for_format (write)

    # read pass no path
    with pytest.raises(ValueError):
        stores.new_for_format(fmt=None, mode='r')

    # write pass path=basedir
    p = tmpdir.mkdir('a').strpath
    fullp = os.path.join(p, stores.STORE_MD_FILENAME)
    d = stores.new_for_format(fmt='npy', path=p, imgshape=(10, 10), imgdtype=np.uint8)
    assert d.full_path == fullp
    assert d.mode == 'w'
    d.close()

    # write pass path=fullpath
    p = tmpdir.mkdir('b').strpath
    fullp = os.path.join(p, stores.STORE_MD_FILENAME)
    d = stores.new_for_format(fmt='npy', path=fullp, imgshape=(10, 10), imgdtype=np.uint8)
    assert d.full_path == fullp
    assert d.mode == 'w'
    d.close()

    # write pass basedir=basedir (API/back compat)
    p = tmpdir.mkdir('c').strpath
    fullp = os.path.join(p, stores.STORE_MD_FILENAME)
    d = stores.new_for_format(fmt='npy', basedir=p, imgshape=(10, 10), imgdtype=np.uint8)
    assert d.full_path == fullp
    assert d.mode == 'w'
    d.close()

    with pytest.raises(ValueError):
        # write pass basedir=fullpath
        p = tmpdir.mkdir('d').strpath
        fullp = os.path.join(p, stores.STORE_MD_FILENAME)
        d = stores.new_for_format(fmt='npy', basedir=fullp, imgshape=(10, 10), imgdtype=np.uint8)


@pytest.mark.parametrize("fmt", ['npy', 'mjpeg', 'avc1/mp4', 'h264/mkv'])
def test_odd_sized(fmt, tmpdir):
    img = np.zeros((199, 199, 3), dtype=np.uint8)
    d = stores.new_for_format(fmt,
                              basedir=tmpdir.strpath,
                              imgshape=img.shape,
                              imgdtype=img.dtype)
    for i in range(10):
        d.add_image(img, i, 0.)
    d.close()

    d = stores.new_for_filename(d.full_path)
    _img, _ = d.get_next_image()

    assert _img.shape == d.image_shape


@pytest.mark.parametrize("fmt", ['npy', 'mjpeg'])
def test_empty(tmpdir, fmt):
    s = stores.new_for_format(fmt,
                              basedir=tmpdir.strpath,
                              imgshape=(512, 512),
                              imgdtype=np.uint8,
                              chunksize=17)
    s.close()

    d = stores.new_for_filename(s.full_path)
    assert np.isnan(d.frame_min)
    assert np.isnan(d.frame_max)
    assert d.frame_count == 0
    assert d.duration == 0
    assert d.chunks == []


NMTW_FNS = [6050, 6055, 6056, 0, 1, 2]
def _non_monotonic_store_with_wrap(fmt, basedir, chunksize, save_index):

    s = stores.new_for_format(fmt,
                              basedir=basedir,
                              imgshape=(512, 512),
                              imgdtype=np.uint8,
                              chunksize=chunksize)

    times = []
    for fn in NMTW_FNS:
        times.append(time.time())
        s.add_image(encode_image(fn, imgsize=512), fn, times[-1])
    s.close(save_index=save_index)

    return s.full_path, times


@pytest.mark.parametrize('chunksize', (2, 100))
@pytest.mark.parametrize("fmt", ['npy', 'mjpeg'])
def test_complex_framenumber(tmpdir, chunksize, fmt):
    s_full_path, times = _non_monotonic_store_with_wrap(fmt, tmpdir.strpath, chunksize, save_index=False)

    d = stores.new_for_filename(s_full_path)
    assert d.frame_min == 0
    assert d.frame_max == 6056
    assert d.frame_count == 6
    assert d.chunks == ([0, 1, 2] if chunksize == 2 else [0])

    for fn, t in zip(NMTW_FNS, times):
        frame, (frame_number, frame_timestamp) = d.get_next_image()
        assert frame.shape == (512, 512)
        assert frame_number == fn
        assert frame_timestamp == t
        assert decode_image(frame) == fn

    with pytest.raises(EOFError):
        d.get_next_image()

    # test some error paths
    with pytest.raises(ValueError):
        d.get_image(42, exact_only=True, frame_index=None)
    with pytest.raises(ValueError):
        d.get_image(None, exact_only=True, frame_index=42)

    # old behaviour
    frame, (frame_number, frame_timestamp) = d.get_image(42, exact_only=False, frame_index=None)
    assert frame_number == 2

    # iter in reversed to make it more interesting
    for fn in reversed(NMTW_FNS):
        frame, (frame_number, frame_timestamp) = d.get_image(fn, exact_only=True, frame_index=None)
        assert frame_number == fn
        assert decode_image(frame) == fn

    for fn in reversed(NMTW_FNS):
        idx = NMTW_FNS.index(fn)
        frame, (frame_number, frame_timestamp) = d.get_image(frame_number=None, exact_only=True,
                                                             frame_index=idx)
        assert frame_number == fn
        assert decode_image(frame) == fn

    # seek to middle and then keep reading
    fn = 6055
    frame, (frame_number, frame_timestamp) = d.get_image(fn, exact_only=True, frame_index=None)
    assert frame_number == fn
    assert decode_image(frame) == fn

    fn = 6056
    frame, (frame_number, frame_timestamp) = d.get_next_image()
    assert frame_number == fn
    assert decode_image(frame) == fn

    fn = 0
    frame, (frame_number, frame_timestamp) = d.get_next_image()
    assert frame_number == fn
    assert decode_image(frame) == fn

    # get nearest frames
    frame, (frame_number, frame_timestamp) = d.get_image(6054, exact_only=False, frame_index=None)
    assert frame_number == 6055
    assert decode_image(frame) == 6055

    frame, (frame_number, frame_timestamp) = d.get_nearest_image(times[0])
    assert frame_number == 6050
    assert decode_image(frame) == 6050
    frame, (frame_number, frame_timestamp) = d.get_nearest_image(times[0] - 1.)
    assert frame_number == 6050
    assert decode_image(frame) == 6050
    frame, (frame_number, frame_timestamp) = d.get_nearest_image(times[0] + (3. * (times[1] - times[0]) / 4.))
    assert frame_number == 6055
    assert decode_image(frame) == 6055


@pytest.mark.parametrize('chunksize', (2, 100))
def test_index(tmpdir, chunksize):
    from imgstore.constants import STORE_INDEX_FILENAME, FRAME_MD
    from imgstore.index import ImgStoreIndex

    basedir = tmpdir.strpath
    s_full_path, times = _non_monotonic_store_with_wrap('npy', basedir, chunksize, save_index=True)

    assert os.path.exists(os.path.join(basedir, STORE_INDEX_FILENAME))

    idx = ImgStoreIndex.new_from_file(os.path.join(basedir, STORE_INDEX_FILENAME))
    assert list(idx.chunks) == ([0, 1, 2] if chunksize == 2 else [0])

    d = stores.new_for_filename(s_full_path)
    assert d.chunks == list(idx.chunks)
    assert d._chunk_n_and_chunk_paths is not None

    for cn in idx.chunks:
        mdd = d._index.get_chunk_metadata(cn)
        mdi = idx.get_chunk_metadata(cn)

        for n in FRAME_MD:
            assert tuple(mdd[n]) == tuple(mdi[n])

    f = stores.new_for_filename(s_full_path, index=idx)
    assert f._index is idx
    assert f._chunk_n_and_chunk_paths is None

    for fn, t in zip(NMTW_FNS, times):
        frame, (frame_number, frame_timestamp) = d.get_next_image()
        assert frame.shape == (512, 512)
        assert frame_number == fn
        assert frame_timestamp == t
        assert decode_image(frame) == fn

    assert f._chunk_n_and_chunk_paths is None
    for _ in f._iter_chunk_n_and_chunk_paths():
        pass
    assert f._chunk_n_and_chunk_paths is not None


@pytest.mark.parametrize("fmt", ['npy', 'mjpeg', 'avc1/mp4', 'h264/mkv'])
@pytest.mark.parametrize("nframes", [2, 3, 4])
@pytest.mark.parametrize("seek", [True, False], ids=['seek', 'noseek'])
def test_max_framenumber_behaviour(fmt, nframes, seek, tmpdir):
    _, path = new_framecode_store(dest=tmpdir.strpath,
                                  frame0=3, nframes=nframes, format=fmt, chunksize=3)

    store = stores.new_for_filename(path, seek=seek)
    assert store.frame_count == nframes

    img, (_fn, _) = store.get_image(frame_number=store.frame_max, exact_only=True)
    assert _fn == store.frame_max
    assert img is not None

    with pytest.raises(EOFError):
        store.get_next_framenumber()

    store.close()
    store = stores.new_for_filename(path, seek=seek)

    img, (_fn, _) = store.get_image(frame_number=store.frame_max - 1, exact_only=True)
    assert decode_image(img[:, :, 0]) == (store.frame_max - 1)
    img, (_fn, _) = store.get_image(frame_number=store.frame_max, exact_only=True)
    assert _fn == store.frame_max
    assert decode_image(img[:, :, 0]) == store.frame_max
    assert store.frame_number == store.frame_max

    with pytest.raises(EOFError):
        store.get_next_framenumber()

    store.close()
    store = stores.new_for_filename(path, seek=seek)

    for i in range(0, store.frame_count):
        img, (_fn, _) = store.get_image(frame_number=None, exact_only=True, frame_index=i)
        assert (_fn - 3) == i
        assert decode_image(img[:, :, 0]) == _fn

    with pytest.raises(ValueError):
        store.get_image(frame_number=None, exact_only=True, frame_index=store.frame_count)


@pytest.mark.parametrize('chunksize', (2, 100))
@pytest.mark.parametrize("fmt", ['npy', 'mjpeg'])
def test_extra_data(tmpdir, chunksize, fmt):

    s = stores.new_for_format(fmt,
                              basedir=tmpdir.strpath,
                              imgshape=(512, 512),
                              imgdtype=np.uint8,
                              chunksize=chunksize)

    frame_extra = {}
    for fn in range(30):
        s.add_image(encode_image(fn, imgsize=512), fn, time.time())

        i = fn + 42
        s.add_extra_data(N=i)
        frame_extra[i] = dict(N=i)

    s.close()

    d = stores.new_for_filename(s.full_path)

    assert d.has_extra_data

    df = d.get_extra_data()
    assert len(df) == len(frame_extra)
    df.set_index('frame_number', inplace=True, verify_integrity=True)
    for _i, _v in frame_extra.items():
        fn = _i - 42
        _row = df.loc[fn]
        assert _row['N'] == _i

    d.close()


@pytest.mark.parametrize("nframes", [0,1])
@pytest.mark.parametrize("fmt", ['npy', 'mjpeg'])
def test_short(tmpdir, nframes, fmt):

    with stores.new_for_format(fmt,
                               basedir=tmpdir.strpath,
                               imgshape=(512, 512),
                               imgdtype=np.uint8,
                               chunksize=13) as s:
        for fn in range(nframes):
            s.add_image(encode_image(fn, imgsize=512), fn, time.time())

    d = stores.new_for_filename(s.full_path)
    assert d.frame_count == nframes

    for i in range(nframes):
        img, (_fn, _) = d.get_image(frame_number=i, exact_only=True)
        assert decode_image(img) == i



import cv2
import numpy as np

FONT_POS_TOP = 0
FONT_POS_BOTTOM = 1
FONT_POS_TOP_LEFT = 2

DEFAULT_FONT = cv2.FONT_HERSHEY_SIMPLEX


# todo: opencv 3.3? added getFontScaleForHeight which can help to auto-scale text


def fit_text(txt, img, pos=FONT_POS_TOP, color=(255, 255, 255), name=cv2.FONT_HERSHEY_SIMPLEX, only_measure=False,
             img_shape=None):
    if pos not in {FONT_POS_BOTTOM, FONT_POS_TOP}:
        raise ValueError('invalid position')

    if (img is None) and (img_shape is None):
        raise ValueError('img or img_shape must be provided')

    def _measure_string(_scale, _thickness, _pad=1):
        # getTextSize under-measures height
        (w, h), _ = cv2.getTextSize(txt, name, _scale, _thickness)
        return _pad + w, _pad + h + (int(_scale + 0.5) * 1)

    tol = 0.90  # text should be 90% of the image width

    if img_shape is None:
        img_shape = img.shape[:2]

    ih, iw = img_shape
    tw, th = iw, ih
    scale = thickness = 1

    # keep bisecting a scale until the text it >= 90% of the image width
    tries = 0
    pct = (tw / float(iw))
    while (tries < 20) and (not ((pct > tol) and (tw < iw))):
        # shitty binary-ish search
        # increase by 100% if smaller, decrease by 25% if larger, so as to approach the limit
        scale *= (2.0 if pct < tol else 0.75)
        thickness = int(max(1, scale // 2))
        tw, th = _measure_string(scale, thickness)
        pct = (tw / float(iw))
        tries += 1

    if pos == FONT_POS_TOP:
        pos = (1, th + 3)
    elif pos == FONT_POS_BOTTOM:
        pos = (1, ih - 3)

    if not only_measure:
        cv2.putText(img, txt, pos, name, scale, color, thickness)

    return scale, thickness


def measure_string(s, name, scale, thickness):
    # getTextSize under-measures height
    (w, h), _ = cv2.getTextSize(s, name, scale, thickness)
    return w, h + (int(scale + 0.5) * 1)


class FontRenderer(object):

    def __init__(self, name, scale, thickness):
        self.name = name
        self.scale = scale
        self.thickness = thickness

    def measure_string(self, s):
        return measure_string(s, self.name, self.scale, self.thickness)

    def write_string(self, s, pos, color, img=None, bgcolor=None):
        """
        write the string onto the image.

        If no image supplied, create one that is cropped to the size of the text, and return it.
        The local origin is bottom left of text string, image coordinates are TL=0,0 BR=W,H

        Args:
            s: string
            pos: (x,y), or FONT_POS_TOP_LEFT
                bottom left coordinate of string
            color: (b,g,r)
            img: array

        Returns: the image written onto
        """
        w = h = None

        if pos == FONT_POS_TOP_LEFT:
            w, h = self.measure_string(s)
            pos = (0, h)

        # make a large enough image
        if img is None:
            if w is None:
                w, h = self.measure_string(s)
            img = np.zeros((h + pos[1] + 1, w + pos[0] + 1, 3), np.uint8)
            pos = pos[0], pos[1] + h

        if bgcolor is not None:
            if w is None:
                w, h = self.measure_string(s)
            x, y = pos
            pt1 = (x, y)
            pt2 = (x + w, y - h)
            cv2.rectangle(img, pt1, pt2, bgcolor, -1)

        cv2.putText(img, s, pos,
                    self.name, self.scale, color, self.thickness)

        return img

    def write_paragraph(self, s, pos, color, img=None):
        w = 0

        if pos == FONT_POS_TOP_LEFT:
            pos = (0, 0)

        x, y = pos
        lines = []
        # measure and layout all lines
        for l in s.split('\n'):
            _w, h = self.measure_string(l)
            w = max(_w, w)
            lines.append((l, (x, y + h)))
            y += int(1.2 * h)

        # make a large enough image
        if img is None:
            img = np.zeros((y, w + x, 3), np.uint8)

        for l, pos in lines:
            self.write_string(l, pos, color, img=img)

        return img


class FontRendererStatusBar(FontRenderer):
    POS_TOP = FONT_POS_TOP
    POS_BOTTOM = FONT_POS_BOTTOM
    POS_TOP_LEFT = FONT_POS_TOP_LEFT

    def __init__(self, name, scale=1, thickness=1):
        super(FontRendererStatusBar, self).__init__(name, scale, thickness)

        self._pos = None
        # x,y,w,h (w,h include padding)
        self._layout_l = None
        self._layout_r = None
        self._layout_box = None

    @classmethod
    def new_for_image(cls, img, font_name=DEFAULT_FONT, pos=POS_BOTTOM, **layout_and_init_args):
        h, w = img.shape[:2]

        # check if all the text can in principle fit...
        txt = ' '  # one space between left and right required
        txt += layout_and_init_args.get('ltext', '')
        txt += layout_and_init_args.get('rtext', '')
        scale = layout_and_init_args.get('scale', 1)
        thickness = layout_and_init_args.get('thickness', 1)
        padh = layout_and_init_args.get('padh', 12)
        _w, _h = measure_string(txt, font_name, scale, thickness)
        if (_w + (2 * padh)) > w:
            layout_and_init_args['scale'], layout_and_init_args['thickness'] = fit_text(txt, img=None,
                                                                                        name=font_name,
                                                                                        only_measure=True,
                                                                                        img_shape=(h, w - (2 * padh)))

        return FontRendererStatusBar.new_for_image_size((w, h),
                                                        font_name,
                                                        pos,
                                                        **layout_and_init_args)

    @classmethod
    def new_for_image_size(cls, frame_size, font_name=DEFAULT_FONT, pos=POS_BOTTOM, **layout_and_init_args):
        init_args = {}
        try:
            init_args['scale'] = layout_and_init_args.pop('scale')
        except KeyError:
            pass
        try:
            init_args['thickness'] = layout_and_init_args.pop('thickness')
        except KeyError:
            pass

        renderer = cls(font_name, **init_args)
        renderer.layout(pos, frame_size, **layout_and_init_args)

        return renderer

    def layout(self, pos, winsize, ltext="", rtext="", padv=4, padh=12):
        w, h = winsize

        ltw, lth = self.measure_string(ltext) if ltext else (0, 0)
        rtw, rth = self.measure_string(rtext) if rtext else (0, 0)

        if pos == FONT_POS_BOTTOM:
            if ltext:
                self._layout_l = (padh, h - padv, ltw, lth)
            if rtext:
                self._layout_r = (w - rtw - padh, h - padv, rtw, rth)
            bh = max(lth + (2 * padv), rth + (2 * padv))
            self._layout_box = (0, h - bh, w, bh)
        elif pos == FONT_POS_TOP:
            if ltext:
                self._layout_l = (padh, lth + padv, ltw, lth)
            if rtext:
                self._layout_r = (w - rtw - padh, rth + padv, rtw, rth)
            self._layout_box = (0, 0, w, max(lth + (2 * padv), rth + (2 * padv)))

    def render(self, img, ltext="", lcolor=(255, 255, 255), rtext="", rcolor=(255, 255, 255), bgcolor=None):
        if bgcolor:
            x, y, w, h = self._layout_box
            pt1 = (x, y)
            pt2 = (x + w, y + h)
            cv2.rectangle(img, pt1, pt2, bgcolor, -1)

        if ltext and self._layout_l:
            cv2.putText(img, ltext,
                        (self._layout_l[0], self._layout_l[1]),
                        self.name, self.scale, lcolor, self.thickness)
        if rtext and self._layout_r:
            cv2.putText(img, rtext,
                        (self._layout_r[0], self._layout_r[1]),
                        self.name, self.scale, rcolor, self.thickness)

    def extract(self, img):
        l, r = None, None
        x, y, w, h = self._layout_l if self._layout_l else (0, 0, 0, 0)
        if w:
            l = img[y - h:y + 1, x:x + w + 1, :]
        x, y, w, h = self._layout_r if self._layout_r else (0, 0, 0, 0)
        if w:
            r = img[y - h:y + 1, x:x + w + 1, :]
        return l, r



import re
import sys
import json
import datetime
import collections

import cv2
import tzlocal

import numpy as np

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    from backports.zoneinfo import ZoneInfo


FourCC = cv2.VideoWriter_fourcc


def _cvt_color(img, code, ensure_copy=True):
    # protect against empty last dimensions
    _is_color = (img.shape[-1] == 3) & (img.ndim == 3)

    if code == cv2.COLOR_GRAY2BGR:
        if _is_color:
            return img.copy() if ensure_copy else img
        else:
            return cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    elif code == cv2.COLOR_BGR2GRAY:
        if _is_color:
            return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        else:
            return img.copy() if ensure_copy else img
    else:
        return ValueError("cvtColor code not understood: %s" % code)


def ensure_grayscale(img, ensure_copy=False):
    return _cvt_color(img, cv2.COLOR_BGR2GRAY, ensure_copy=ensure_copy)


def ensure_color(img, ensure_copy=False):
    return _cvt_color(img, cv2.COLOR_GRAY2BGR, ensure_copy=ensure_copy)


class ImageCodecProcessor(object):
    """
    Allows converting a number of image encodings (Bayer, YUV, etc)
    to OpenCV native format (BGR).

    Because I think OpenCV disagrees about what to name bayer formats
    as compared to other libraries, all encodings are prefixed with
    'cv_'
    """

    # https://regex101.com/r/pL2mY2/1
    CV2_CODEC_RE = re.compile("^COLOR_(?P<codec>[A-Za-z0-9_]+)2BGR($|(?P<method>_[A-Za-z0-9]*))")

    def __init__(self):
        self._default_code = None
        self._default_enum = None

        self._cv2_code_to_enum = {}
        self._cv2_enum_to_code = {}
        # an extra private mapping because OpenCV has many enums mapping to the same value
        self.__cv2_enum_to_name = collections.defaultdict(list)
        for name in dir(cv2):
            m = self.CV2_CODEC_RE.match(name)
            if m:
                enum = int(getattr(cv2, name))
                self.__cv2_enum_to_name[enum].append(name)

                parts = m.groupdict()
                code = 'cv_%s' % parts['codec'].lower().replace('_', '')

                meth = parts.get('method', None)
                if meth:
                    if 'bayer' in code:
                        # cv includes different bayer methods, ignore them
                        continue
                    elif meth == 'full':
                        # fixme: not sure what this is
                        continue

                    code += '_%s' % parts['method'].lower().replace('_', '')

                self._cv2_code_to_enum[code] = enum
                self._cv2_enum_to_code[enum] = code

        self.autoconvert = lambda x: x

        # pprint.pprint(self._cv2_code_to_enum)
        # pprint.pprint(self._cv2_enum_to_code)

    def __repr__(self):
        return "<ImageCodecProcessor(%s)>" % self._default_code

    @classmethod
    def from_pylon_format(cls, s):
        # pylon disagrees with opencv (B & R are always switched)
        if s in {'BayerBG', 'Bayer_BG'}:
            code = 'cv_bayerrg'
        elif s in {'BayerRG', 'Bayer_RG'}:
            code = 'cv_bayerbg'
        elif s in {'BayerGB', 'Bayer_GB'}:
            code = 'cv_bayergr'
        elif (not s) or (s == 'None'):
            code = None
        else:
            raise ValueError('unknown pylon format: %r' % s)

        obj = cls()
        obj.set_default_code(code)
        return obj

    @classmethod
    def from_cv2_enum(cls, e):
        obj = cls()
        obj.set_default_code(obj.cv2_enum_to_code(e))
        return obj

    @property
    def encoding(self):
        return self._default_code or ''

    def _autodecode_cvtcolor(self, img):
        return cv2.cvtColor(img, self._default_enum)

    def set_default_code(self, code):
        if not code:
            self._default_code = None
            self._default_enum = None
            self.autoconvert = lambda x: x
        else:
            if not self.check_code(code):
                raise ValueError('unknown image encoding:' % code)
            self._default_code = code
            self._default_enum = self._cv2_code_to_enum[code]
            self.autoconvert = self._autodecode_cvtcolor

    def check_code(self, code):
        return code in self._cv2_code_to_enum

    def cv2_enum_to_code(self, enum):
        return self._cv2_enum_to_code[enum]

    def convert_to_bgr(self, img, code=None, method_enum=None):
        if code is None:
            code = self._default_code
        if method_enum is None:
            method_enum = self._cv2_code_to_enum[code]
        if code is None:
            raise ValueError('encoding must be specified')
        return cv2.cvtColor(img, method_enum)


class JsonCustomEncoder(json.JSONEncoder):
    def default(self, obj):
        # noinspection PyUnresolvedReferences
        if isinstance(obj, (np.ndarray, np.number)):
            return obj.tolist()
        elif isinstance(obj, (complex, np.complex)):
            return [obj.real, obj.imag]
        elif isinstance(obj, set):
            return list(obj)
        elif isinstance(obj, bytes):  # pragma: py3
            return obj.decode()
        elif isinstance(obj, datetime.datetime):
            return obj.isoformat()

        return json.JSONEncoder.default(self, obj)


def motif_extra_data_json_to_df(store, path):
    import pandas as pd

    with open(path, 'rt') as f:
        records = json.load(f)
        df = pd.DataFrame(records)

        if not df.empty:
            df = df[df['frame_number'] >= 0]

        if not df.empty:

            if 'sensor_time' in df.columns:
                by = ['frame_index','sensor_time']
            else:
                by = ['frame_index']

            try:
                return df.sort_values(by=by, ignore_index=True)
            except TypeError:
                return df.sort_values(by=by).reset_index(drop=True)


def motif_get_parse_true_fps(store, default=25.0, hwardware_only=False):
    md = store.user_metadata
    if 'hwframerate' in md:
        return float(md['hwframerate'])
    elif 'motifptpframerate' in md:
        return float(md['motifptpframerate'])
    elif (not hwardware_only) and \
            (md.get('acquisitionframerate') and (md.get('acquisitionframerateenable', False) is True)):
        return float(md['acquisitionframerate'])
    elif default is None:
        return 1.0 / np.median(np.diff(store._get_chunk_metadata(0)['frame_time']))
    return default


def motif_extra_data_h5_attrs(path):
    import h5py

    attrs = {}
    with h5py.File(path, 'r') as f:
        attrs['_datasets'] = [s.strip() for s in f.attrs['datasets'].decode('ascii').split(',')]
        for g in f.keys():
            attrs[g] = dict(f[g].attrs)

    return attrs


def motif_extra_data_h5_to_df(store, path):
    import h5py
    import pandas as pd

    def _attr_string(_s):
        try:
            return _s.decode('ascii')
        except AttributeError:
            return _s

    with h5py.File(path, 'r') as f:
        dat = {}

        # the camera information is stored in an array with compound datatype
        camera = np.asarray(f['camera'])

        # recording can be stopped before chunk is full, so trim away rows in the store that
        # were pre-allocated, but not recorded. pre-allocated but un-used frames are indicated with
        # a framenumber < 0
        mask = camera['frame_number'] >= 0

        # motif stores the names of datasets in a root attribute
        datasets = [s.strip() for s in _attr_string(f.attrs['datasets']).split(',')]

        for dsname in datasets:
            ds = f[dsname]
            col_names = [s.strip() for s in _attr_string(ds.attrs['column_names']).split(',')]

            # trim the array
            arr = ds[..., mask]

            assert len(col_names) == arr.shape[0]

            for col, col_name in enumerate(col_names):
                dat[col_name] = arr[col]

        # also add the camera sync information
        for cam_md_name in camera.dtype.names:
            dat[cam_md_name] = camera[cam_md_name][mask]

        # and the sample delays
        dat['sample_delay'] = np.asarray(f['sample_delay'])[mask]

        return pd.DataFrame(dat)


def motif_get_recording_timezone(store):
    _, tz = store.created
    if store.user_metadata and (store.user_metadata.get('timezone')):
        try:
            return ZoneInfo(store.user_metadata['timezone'])
        except:
            pass
    return tz


def get_local_timezone():
    # tzlocal is probbably the worst API I have ever used at maintaining backwards
    # compatibility.
    try:
        return tzlocal.get_localzone().zone
    except AttributeError:
        return tzlocal.get_localzone_name()


def get_local_timezone_zoneinfo():
    return ZoneInfo(get_local_timezone())




#!/usr/bin/env python

# Authors: John Stowers <john@loopbio.com>, Santi Villalba <santi@loopbio.com>
# Licence: BSD 3 clause

import os.path as op
from setuptools import setup, find_packages

this_directory = op.abspath(op.dirname(__file__))
with open(op.join(this_directory, 'README.md'), 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='imgstore',
    license='BSD 3 clause',
    description='IMGStore houses your video frames',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    version='0.4.0',
    url='https://github.com/loopbio/imgstore',
    author='John Stowers, Santi Villalba',
    author_email='john@loopbio.com, santi@loopbio.com',
    python_requires='>=3.6',
    packages=find_packages(),
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    install_requires=[
        'numpy',
        'pandas',
        'pyyaml',
        'backports.zoneinfo; python_version < "3.9.0"',
        'tzlocal',
        'python-dateutil',
    ],
    tests_require=[
        'pytest',
        'pytest-cov',
        'pytest-pep8'
    ],
    extras_require={
        'bloscpack': ['bloscpack']
    },
    entry_points={
        'console_scripts': [
            'imgstore-view = imgstore.apps:main_viewer',
            'imgstore-save = imgstore.apps:main_saver',
            'imgstore-test = imgstore.apps:main_test',
        ]
    },
)



import itertools
import functools

import numpy as np


class StoreAligner(object):
    MISSING_POLICY_DROP = 'drop'
    MISSING_POLICY_HOLD = 'zoh'
    MISSING_POLICY_BLACK = 'black'

    def __init__(self, *stores):
        self._fns = []
        self._missing_policy = None
        self._stores = stores
        sizes = {s.image_shape for s in stores}
        assert len(sizes) == 1
        self._store_shape = sizes.pop()

    def extract_common_frames(self, policy):
        assert policy in (StoreAligner.MISSING_POLICY_DROP,
                          StoreAligner.MISSING_POLICY_HOLD,
                          StoreAligner.MISSING_POLICY_BLACK)

        fns_arr = [s.get_frame_metadata()['frame_number'] for s in self._stores]
        if self._missing_policy == StoreAligner.MISSING_POLICY_DROP:
            self._fns = functools.reduce(np.intersect1d, fns_arr)
        else:
            fns_all = np.fromiter(itertools.chain.from_iterable(fns_arr), dtype=int)
            self._fns = np.unique(fns_all)

        self._missing_policy = policy
        return self._fns

    def iter_imgs(self, return_times=False):
        assert len(self._fns)
        assert self._missing_policy is not None

        black_img = np.zeros(self._store_shape, dtype=np.uint8)
        last_imgs = [black_img for _ in range(len(self._stores))]

        for fn in self._fns:
            imgs = []
            ts = []
            for i, store in enumerate(self._stores):
                try:
                    img, (_, t) = store.get_image(frame_number=fn, exact_only=True)
                    last_imgs[i] = img
                    imgs.append(img)
                    ts.append(t)
                except ValueError:
                    print('store %d missing frame %s' % (i, fn))
                    ts.append(np.nan)
                    if self._missing_policy == StoreAligner.MISSING_POLICY_HOLD:
                        imgs.append(last_imgs[i])
                    elif self._missing_policy == StoreAligner.MISSING_POLICY_BLACK:
                        imgs.append(black_img)
                    elif self._missing_policy == StoreAligner.MISSING_POLICY_DROP:
                        continue
                    else:
                        raise NotImplementedError

            if len(imgs) != len(self._stores):
                continue

            if return_times:
                yield fn, imgs, ts
            else:
                yield fn, imgs
